#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for common methods used by generic framework

Author: A Lee
"""
import pytest
from types import MethodType

from pyomo.environ import Block, ConcreteModel, units as pyunits, Var
from pyomo.common.config import ConfigBlock, ConfigValue

from idaes.core import declare_process_block_class
from idaes.core import Component, LiquidPhase, SolidPhase, VaporPhase, PhaseType as PT
from idaes.models.properties.modular_properties.base.utility import (
    GenericPropertyPackageError,
    get_method,
    get_phase_method,
    get_component_object,
    get_bounds_from_config,
    get_concentration_term,
    ConcentrationForm,
    identify_VL_component_list,
    estimate_Pbub,
    estimate_Pdew,
    estimate_Tbub,
    estimate_Tdew,
)
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterData,
)
from idaes.models.properties.modular_properties.base.generic_reaction import rxn_config
from idaes.core.util.exceptions import ConfigurationError, PropertyPackageError
from idaes.core.util.misc import add_object_reference
from idaes.models.properties.modular_properties.base.tests.dummy_eos import DummyEoS
from idaes.models.properties.modular_properties.phase_equil.henry import (
    ConstantH,
    HenryType,
)
from idaes.models.properties.modular_properties.eos.ceos import Cubic, CubicType
from idaes.models.properties.modular_properties.phase_equil import SmoothVLE
from idaes.models.properties.modular_properties.phase_equil.bubble_dew import (
    LogBubbleDew,
)
from idaes.models.properties.modular_properties.phase_equil.forms import log_fugacity
from idaes.models.properties.modular_properties.pure import NIST
from idaes.models.properties.modular_properties.state_definitions import FTPx


@pytest.fixture
def frame():
    m = Block(concrete=True)
    m.params = Block()
    m.params.config = ConfigBlock()
    m.params.config.declare("test_arg", ConfigValue())
    m.params.config.declare("state_bounds", ConfigValue())

    def get_component(self, comp):
        return getattr(self, comp)

    m.params.get_component = MethodType(get_component, m.params)
    m.params.get_phase = MethodType(get_component, m.params)

    m.params.comp = Block()
    m.params.comp.config = ConfigBlock()
    m.params.comp.config.declare("test_arg_2", ConfigValue())

    return m


@pytest.mark.unit
def test_generic_property_package_error():
    with pytest.raises(
        PropertyPackageError,
        match="Generic Property Package instance block called for "
        "prop, but was not provided with a method "
        "for this property. Please add a method for this property "
        "in the property parameter configuration.",
    ):
        raise GenericPropertyPackageError("block", "prop")


@pytest.mark.unit
def test_get_component_object(frame):
    assert get_component_object(frame, "comp") is frame.params.comp


class TestGetMethod:
    @pytest.mark.unit
    def test_get_method_invalid_name(self, frame):
        with pytest.raises(
            AttributeError,
            match="ScalarBlock Generic Property Package called for "
            "invalid configuration option foo. Please contact the "
            "developer of the property package.",
        ):
            get_method(frame, "foo")

    @pytest.mark.unit
    def test_get_method_none(self, frame):
        with pytest.raises(
            GenericPropertyPackageError,
            match="Generic Property Package instance ScalarBlock "
            "called for test_arg, but was not provided with a "
            "method for this property. Please add a method for "
            "this property in the property parameter "
            "configuration.",
        ):
            get_method(frame, "test_arg")

    @pytest.mark.unit
    def test_get_method_not_callable(self, frame):
        frame.params.config.test_arg = "foo"
        with pytest.raises(
            ConfigurationError,
            match="ScalarBlock Generic Property Package received "
            "invalid value for argument test_arg. Value must be a "
            "method, a class with a method named expression or a "
            "module containing one of the previous.",
        ):
            get_method(frame, "test_arg")

    @pytest.mark.unit
    def test_get_method_simple(self, frame):
        def test_arg():
            return "bar"

        frame.params.config.test_arg = test_arg

        mthd = get_method(frame, "test_arg")
        assert mthd() == "bar"

    @pytest.mark.unit
    def test_get_method_class_w_method(self, frame):
        class TestClass:
            def test_arg():
                return "bar"

        frame.params.config.test_arg = TestClass

        mthd = get_method(frame, "test_arg")
        assert mthd() == "bar"

    @pytest.mark.unit
    def test_get_method_class_w_return_expression(self, frame):
        class TestClass:
            def return_expression(*args, **kwargs):
                return "bar"

        frame.params.config.test_arg = TestClass

        mthd = get_method(frame, "test_arg")
        assert mthd() == "bar"

    @pytest.mark.unit
    def test_get_method_phase(self, frame):
        def test_arg():
            return "bar"

        frame.params.config.test_arg = {"test_phase": test_arg}

        mthd = get_method(frame, "test_arg", phase="test_phase")
        assert mthd() == "bar"

    @pytest.mark.unit
    def test_get_method_comp(self, frame):
        def test_arg():
            return "bar"

        frame.params.comp.config.test_arg_2 = test_arg

        mthd = get_method(frame, "test_arg_2", comp="comp")
        assert mthd() == "bar"

    @pytest.mark.unit
    def test_get_method_phase_comp(self, frame):
        def test_arg():
            return "bar"

        frame.params.comp.config.test_arg_2 = {"test_phase": test_arg}

        mthd = get_method(frame, "test_arg_2", comp="comp", phase="test_phase")
        assert mthd() == "bar"


class TestGetPhaseMethod:
    @pytest.mark.unit
    def test_get_phase_method_invalid_name(self, frame):
        with pytest.raises(
            AttributeError,
            match="ScalarBlock Generic Property Package called for "
            "invalid configuration option foo. Please contact the "
            "developer of the property package.",
        ):
            get_phase_method(frame, "foo", "comp")

    @pytest.mark.unit
    def test_get_phase_method_none(self, frame):
        with pytest.raises(
            GenericPropertyPackageError,
            match="Generic Property Package instance ScalarBlock "
            "called for test_arg_2, but was not provided with a "
            "method for this property. Please add a method for "
            "this property in the property parameter "
            "configuration.",
        ):
            get_phase_method(frame, "test_arg_2", "comp")

    @pytest.mark.unit
    def test_get_phase_method_not_callable(self, frame):
        frame.params.comp.config.test_arg_2 = "foo"
        with pytest.raises(
            ConfigurationError,
            match="ScalarBlock Generic Property Package received "
            "invalid value for argument test_arg_2. Value must be a "
            "method, a class with a method named expression or a "
            "module containing one of the previous.",
        ):
            get_phase_method(frame, "test_arg_2", "comp")

    @pytest.mark.unit
    def test_get_phase_method_simple(self, frame):
        def test_arg():
            return "bar"

        frame.params.comp.config.test_arg_2 = test_arg

        mthd = get_phase_method(frame, "test_arg_2", "comp")
        assert mthd() == "bar"

    @pytest.mark.unit
    def test_get_phase_method_class_w_method(self, frame):
        class TestClass:
            def test_arg_2():
                return "bar"

        frame.params.comp.config.test_arg_2 = TestClass

        mthd = get_phase_method(frame, "test_arg_2", "comp")
        assert mthd() == "bar"

    @pytest.mark.unit
    def test_get_phase_method_class_w_return_expression(self, frame):
        class TestClass:
            def return_expression():
                return "bar"

        frame.params.comp.config.test_arg_2 = TestClass

        mthd = get_phase_method(frame, "test_arg_2", "comp")
        assert mthd() == "bar"


class TestGetBoundsFromConfig:
    @pytest.mark.unit
    def test_no_state_bounds(self, frame):
        bounds, value = get_bounds_from_config(frame, "foo", "bar")

        assert bounds == (None, None)
        assert value is None

    @pytest.mark.unit
    def test_no_state_bounds_3(self, frame):
        frame.params.config.state_bounds = {"test_state": (1, 2, 3)}
        bounds, value = get_bounds_from_config(frame, "test_state", "bar")

        assert bounds == (1, 3)
        assert value == 2

    @pytest.mark.unit
    def test_no_state_bounds_4(self, frame):
        frame.params.config.state_bounds = {"test_state": (1, 2, 3, pyunits.km)}
        bounds, value = get_bounds_from_config(frame, "test_state", pyunits.m)

        assert bounds == (1000, 3000)
        assert value == 2000


class TestGetConcentrationTerm:
    @pytest.fixture
    def frame(self):
        m = Block(concrete=True)

        m.params = Block()

        m.params.config = ConfigBlock()

        m.params.config.declare(
            "rate_reactions", ConfigBlock(implicit=True, implicit_domain=rxn_config)
        )
        m.params.config.declare(
            "equilibrium_reactions",
            ConfigBlock(implicit=True, implicit_domain=rxn_config),
        )
        m.params.config.declare(
            "inherent_reactions", ConfigBlock(implicit=True, implicit_domain=rxn_config)
        )

        add_object_reference(m, "state_ref", m)

        m.conc_mol_phase_comp = Var()
        m.act_phase_comp = Var()
        m.molality_phase_comp = Var()
        m.mole_frac_phase_comp = Var()
        m.mass_frac_phase_comp = Var()
        m.pressure_phase_comp = Var()

        m.log_conc_mol_phase_comp = Var()
        m.log_act_phase_comp = Var()
        m.log_molality_phase_comp = Var()
        m.log_mole_frac_phase_comp = Var()
        m.log_mass_frac_phase_comp = Var()
        m.log_pressure_phase_comp = Var()

        return m

    @pytest.mark.unit
    def test_rate_molarity(self, frame):
        frame.params.config.rate_reactions["r1"] = rxn_config
        frame.params.config.rate_reactions["r1"].concentration_form = (
            ConcentrationForm.molarity
        )

        assert get_concentration_term(frame, "r1") is frame.conc_mol_phase_comp
        assert (
            get_concentration_term(frame, "r1", log=True)
            is frame.log_conc_mol_phase_comp
        )

    @pytest.mark.unit
    def test_rate_activity(self, frame):
        frame.params.config.rate_reactions["r1"] = rxn_config
        frame.params.config.rate_reactions["r1"].concentration_form = (
            ConcentrationForm.activity
        )

        assert get_concentration_term(frame, "r1") is frame.act_phase_comp
        assert get_concentration_term(frame, "r1", log=True) is frame.log_act_phase_comp

    @pytest.mark.unit
    def test_rate_molality(self, frame):
        frame.params.config.rate_reactions["r1"] = rxn_config
        frame.params.config.rate_reactions["r1"].concentration_form = (
            ConcentrationForm.molality
        )

        assert get_concentration_term(frame, "r1") is frame.molality_phase_comp
        assert (
            get_concentration_term(frame, "r1", log=True)
            is frame.log_molality_phase_comp
        )

    @pytest.mark.unit
    def test_rate_mole_frac(self, frame):
        frame.params.config.rate_reactions["r1"] = rxn_config
        frame.params.config.rate_reactions["r1"].concentration_form = (
            ConcentrationForm.moleFraction
        )

        assert get_concentration_term(frame, "r1") is frame.mole_frac_phase_comp
        assert (
            get_concentration_term(frame, "r1", log=True)
            is frame.log_mole_frac_phase_comp
        )

    @pytest.mark.unit
    def test_rate_mass_frac(self, frame):
        frame.params.config.rate_reactions["r1"] = rxn_config
        frame.params.config.rate_reactions["r1"].concentration_form = (
            ConcentrationForm.massFraction
        )

        assert get_concentration_term(frame, "r1") is frame.mass_frac_phase_comp
        assert (
            get_concentration_term(frame, "r1", log=True)
            is frame.log_mass_frac_phase_comp
        )

    @pytest.mark.unit
    def test_rate_partial_pressure(self, frame):
        frame.params.config.rate_reactions["r1"] = rxn_config
        frame.params.config.rate_reactions["r1"].concentration_form = (
            ConcentrationForm.partialPressure
        )

        assert get_concentration_term(frame, "r1") is frame.pressure_phase_comp
        assert (
            get_concentration_term(frame, "r1", log=True)
            is frame.log_pressure_phase_comp
        )

    @pytest.mark.unit
    def test_equilibrium_molarity(self, frame):
        frame.params.config.equilibrium_reactions["e1"] = rxn_config
        frame.params.config.equilibrium_reactions["e1"].concentration_form = (
            ConcentrationForm.molarity
        )

        assert get_concentration_term(frame, "e1") is frame.conc_mol_phase_comp
        assert (
            get_concentration_term(frame, "e1", log=True)
            is frame.log_conc_mol_phase_comp
        )

    @pytest.mark.unit
    def test_equilibrium_activity(self, frame):
        frame.params.config.equilibrium_reactions["e1"] = rxn_config
        frame.params.config.equilibrium_reactions["e1"].concentration_form = (
            ConcentrationForm.activity
        )

        assert get_concentration_term(frame, "e1") is frame.act_phase_comp
        assert get_concentration_term(frame, "e1", log=True) is frame.log_act_phase_comp

    @pytest.mark.unit
    def test_equilibrium_molality(self, frame):
        frame.params.config.equilibrium_reactions["e1"] = rxn_config
        frame.params.config.equilibrium_reactions["e1"].concentration_form = (
            ConcentrationForm.molality
        )

        assert get_concentration_term(frame, "e1") is frame.molality_phase_comp
        assert (
            get_concentration_term(frame, "e1", log=True)
            is frame.log_molality_phase_comp
        )

    @pytest.mark.unit
    def test_equilibrium_mole_frac(self, frame):
        frame.params.config.equilibrium_reactions["e1"] = rxn_config
        frame.params.config.equilibrium_reactions["e1"].concentration_form = (
            ConcentrationForm.moleFraction
        )

        assert get_concentration_term(frame, "e1") is frame.mole_frac_phase_comp
        assert (
            get_concentration_term(frame, "e1", log=True)
            is frame.log_mole_frac_phase_comp
        )

    @pytest.mark.unit
    def test_equilibrium_mass_frac(self, frame):
        frame.params.config.equilibrium_reactions["e1"] = rxn_config
        frame.params.config.equilibrium_reactions["e1"].concentration_form = (
            ConcentrationForm.massFraction
        )

        assert get_concentration_term(frame, "e1") is frame.mass_frac_phase_comp
        assert (
            get_concentration_term(frame, "e1", log=True)
            is frame.log_mass_frac_phase_comp
        )

    @pytest.mark.unit
    def test_equilibrium_partial_pressure(self, frame):
        frame.params.config.equilibrium_reactions["e1"] = rxn_config
        frame.params.config.equilibrium_reactions["e1"].concentration_form = (
            ConcentrationForm.partialPressure
        )

        assert get_concentration_term(frame, "e1") is frame.pressure_phase_comp
        assert (
            get_concentration_term(frame, "e1", log=True)
            is frame.log_pressure_phase_comp
        )

    @pytest.fixture
    def frame2(self):
        m = Block(concrete=True)

        m.params = Block()

        m.params.config = ConfigBlock()

        m.params.config.declare(
            "inherent_reactions", ConfigBlock(implicit=True, implicit_domain=rxn_config)
        )

        add_object_reference(m, "state_ref", m)

        m.conc_mol_phase_comp = Var()
        m.act_phase_comp = Var()
        m.molality_phase_comp = Var()
        m.mole_frac_phase_comp = Var()
        m.mass_frac_phase_comp = Var()
        m.pressure_phase_comp = Var()

        m.log_conc_mol_phase_comp = Var()
        m.log_act_phase_comp = Var()
        m.log_molality_phase_comp = Var()
        m.log_mole_frac_phase_comp = Var()
        m.log_mass_frac_phase_comp = Var()
        m.log_pressure_phase_comp = Var()

        return m

    @pytest.mark.unit
    def test_inherent_molarity(self, frame2):
        frame2.params.config.inherent_reactions["i1"] = rxn_config
        frame2.params.config.inherent_reactions["i1"].concentration_form = (
            ConcentrationForm.molarity
        )

        assert get_concentration_term(frame2, "i1") is frame2.conc_mol_phase_comp
        assert (
            get_concentration_term(frame2, "i1", log=True)
            is frame2.log_conc_mol_phase_comp
        )

    @pytest.mark.unit
    def test_inherent_activity(self, frame2):
        frame2.params.config.inherent_reactions["i1"] = rxn_config
        frame2.params.config.inherent_reactions["i1"].concentration_form = (
            ConcentrationForm.activity
        )

        assert get_concentration_term(frame2, "i1") is frame2.act_phase_comp
        assert (
            get_concentration_term(frame2, "i1", log=True) is frame2.log_act_phase_comp
        )

    @pytest.mark.unit
    def test_inherent_molality(self, frame2):
        frame2.params.config.inherent_reactions["i1"] = rxn_config
        frame2.params.config.inherent_reactions["i1"].concentration_form = (
            ConcentrationForm.molality
        )

        assert get_concentration_term(frame2, "i1") is frame2.molality_phase_comp
        assert (
            get_concentration_term(frame2, "i1", log=True)
            is frame2.log_molality_phase_comp
        )

    @pytest.mark.unit
    def test_inherent_mole_frac(self, frame2):
        frame2.params.config.inherent_reactions["i1"] = rxn_config
        frame2.params.config.inherent_reactions["i1"].concentration_form = (
            ConcentrationForm.moleFraction
        )

        assert get_concentration_term(frame2, "i1") is frame2.mole_frac_phase_comp
        assert (
            get_concentration_term(frame2, "i1", log=True)
            is frame2.log_mole_frac_phase_comp
        )

    @pytest.mark.unit
    def test_inherent_mass_frac(self, frame2):
        frame2.params.config.inherent_reactions["i1"] = rxn_config
        frame2.params.config.inherent_reactions["i1"].concentration_form = (
            ConcentrationForm.massFraction
        )

        assert get_concentration_term(frame2, "i1") is frame2.mass_frac_phase_comp
        assert (
            get_concentration_term(frame2, "i1", log=True)
            is frame2.log_mass_frac_phase_comp
        )

    @pytest.mark.unit
    def test_inherent_partial_pressure(self, frame2):
        frame2.params.config.inherent_reactions["i1"] = rxn_config
        frame2.params.config.inherent_reactions["i1"].concentration_form = (
            ConcentrationForm.partialPressure
        )

        assert get_concentration_term(frame2, "i1") is frame2.pressure_phase_comp
        assert (
            get_concentration_term(frame2, "i1", log=True)
            is frame2.log_pressure_phase_comp
        )


class TestIdentifyVLComponentList:
    # Declare a base units dict to save code later
    base_units = {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    }

    # Dummy methods for properties
    def set_metadata(self, b):
        pass

    def define_state(self, b):
        pass

    @declare_process_block_class("DummyParameterBlock")
    class DummyParameterData(GenericParameterData):
        def configure(self):
            self.configured = True

    @pytest.fixture()
    def frame(self):
        m = ConcreteModel()
        m.params = DummyParameterBlock(
            components={
                "a": {},
                "b": {},
                "c": {},
            },
            phases={
                "p1": {
                    "type": LiquidPhase,
                    "equation_of_state": DummyEoS,
                },
                "p2": {
                    "type": VaporPhase,
                    "equation_of_state": DummyEoS,
                },
                "p3": {
                    "type": SolidPhase,
                    "equation_of_state": DummyEoS,
                },
            },
            state_definition=self,
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units=self.base_units,
        )

        m.props = m.params.build_state_block([1], defined_state=False)

        return m

    @pytest.mark.unit
    def test_invalid_VL_pair(self, frame):
        with pytest.raises(
            PropertyPackageError,
            match="Phase pair p1-p3 was identified as being a VLE pair, "
            "however p1 is liquid but p3 is not vapor.",
        ):
            identify_VL_component_list(frame.props[1], ("p1", "p3"))

        with pytest.raises(
            PropertyPackageError,
            match="Phase pair p2-p3 was identified as being a VLE pair, "
            "however neither p2 nor p3 is liquid.",
        ):
            identify_VL_component_list(frame.props[1], ("p2", "p3"))

    @pytest.mark.unit
    def test_all_components(self, frame):
        l_phase, v_phase, vl_comps, henry_comps, l_only_comps, v_only_comps = (
            identify_VL_component_list(frame.props[1], ("p1", "p2"))
        )

        assert l_phase == "p1"
        assert v_phase == "p2"
        assert vl_comps == ["a", "b", "c"]
        assert henry_comps == []
        assert l_only_comps == []
        assert v_only_comps == []

    @pytest.mark.unit
    def test_all_types_components(self):
        m = ConcreteModel()
        m.params = DummyParameterBlock(
            components={
                "a": {},
                "b": {
                    "valid_phase_types": PT.liquidPhase,
                },
                "c": {
                    "valid_phase_types": PT.vaporPhase,
                },
                "d": {
                    "valid_phase_types": PT.solidPhase,
                },
                "e": {
                    "parameter_data": {"henry_ref": {"p1": 86}},
                    "henry_component": {
                        "p1": {"method": ConstantH, "type": HenryType.Kpx}
                    },
                },
            },
            phases={
                "p1": {
                    "type": LiquidPhase,
                    "equation_of_state": DummyEoS,
                },
                "p2": {
                    "type": VaporPhase,
                    "equation_of_state": DummyEoS,
                },
                "p3": {
                    "type": SolidPhase,
                    "equation_of_state": DummyEoS,
                },
            },
            state_definition=self,
            pressure_ref=100000.0,
            temperature_ref=300,
            base_units=self.base_units,
        )

        m.props = m.params.build_state_block([1], defined_state=False)

        l_phase, v_phase, vl_comps, henry_comps, l_only_comps, v_only_comps = (
            identify_VL_component_list(m.props[1], ("p1", "p2"))
        )
        assert l_phase == "p1"
        assert v_phase == "p2"
        assert vl_comps == ["a"]
        assert henry_comps == ["e"]
        assert l_only_comps == ["b"]
        assert v_only_comps == ["c"]


# Property configuration for pure water to use in bubble and dew point tests
configuration = {
    # Specifying components
    "components": {
        "H2O": {
            "type": Component,
            "pressure_sat_comp": NIST,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "pressure_crit": (220.6e5, pyunits.Pa),
                "temperature_crit": (647, pyunits.K),
                "omega": 0.344,
                "pressure_sat_comp_coeff": {
                    "A": (3.55959, None),
                    "B": (643.748, pyunits.K),
                    "C": (-198.043, pyunits.K),
                },
            },
        },
    },
    # Specifying phases
    "phases": {
        "Liq": {
            "type": LiquidPhase,
            "equation_of_state": Cubic,
            "equation_of_state_options": {"type": CubicType.PR},
        },
        "Vap": {
            "type": VaporPhase,
            "equation_of_state": Cubic,
            "equation_of_state_options": {"type": CubicType.PR},
        },
    },
    # Set base units of measurement
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    # Specifying state definition
    "state_definition": FTPx,
    "state_bounds": {
        "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
        "temperature": (273.15, 300, 500, pyunits.K),
        "pressure": (5e4, 1e5, 1e6, pyunits.Pa),
    },
    "pressure_ref": (101325, pyunits.Pa),
    "temperature_ref": (298.15, pyunits.K),
    # Defining phase equilibria
    "phases_in_equilibrium": [("Vap", "Liq")],
    "phase_equilibrium_state": {("Vap", "Liq"): SmoothVLE},
    "bubble_dew_method": LogBubbleDew,
    "parameter_data": {
        "PR_kappa": {
            ("H2O", "H2O"): 0.000,
        }
    },
}


class TestBubbleDewPoints:
    @pytest.fixture
    def model(self):
        m = ConcreteModel()
        m.params = DummyParameterBlock(
            **configuration,
        )

        m.props = m.params.build_state_block([1], defined_state=False)

        return m

    @pytest.mark.unit
    def test_bubble_temperature(self, model):
        # Test bubble temperature at atmospheric pressure
        model.props[1].pressure.set_value(101325)
        Tbub = estimate_Tbub(model.props[1], pyunits.K, ["H2O"], [], "Liq")

        # Expected value = 379.1828 from parameters used
        assert Tbub == pytest.approx(379.1828, rel=1e-6)

    @pytest.mark.unit
    def test_dew_temperature(self, model):
        # Test dew temperature at atmospheric pressure
        model.props[1].pressure.set_value(101325)
        Tdew = estimate_Tdew(model.props[1], pyunits.K, ["H2O"], [], "Liq")

        # Expected value = 379.1828 from parameters used
        assert Tdew == pytest.approx(379.1828, rel=1e-6)

    @pytest.mark.unit
    def test_bubble_pressure(self, model):
        # Test bubble pressure at 100C
        model.props[1].temperature.set_value(373.15)
        Pbub = estimate_Pbub(model.props[1], ["H2O"], [], "Liq")

        # Expected value = 76432.45 from parameters used
        assert Pbub == pytest.approx(76432.45, rel=1e-6)

    @pytest.mark.unit
    def test_dew_pressure(self, model):
        # Test dew pressure at 100C
        model.props[1].temperature.set_value(373.15)
        Pdew = estimate_Pdew(model.props[1], ["H2O"], [], "Liq")

        # Expected value = 76432.45 from parameters used
        assert Pdew == pytest.approx(76432.45, rel=1e-6)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for equilibrium reactor unit model.
Authors: Andrew Lee
"""

from math import exp
import pytest

from pyomo.environ import (
    assert_optimal_termination,
    ComponentMap,
    ConcreteModel,
    Suffix,
    TransformationFactory,
    units,
    value,
)

from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
)
from idaes.models.unit_models.equilibrium_reactor import (
    EquilibriumReactor,
    EquilibriumReactorScaler,
)
from idaes.models.properties.examples.saponification_thermo import (
    SaponificationParameterBlock,
)
from idaes.models.properties.examples.saponification_reactions import (
    SaponificationReactionParameterBlock,
)
from idaes.core.util.model_statistics import (
    number_variables,
    number_total_constraints,
    number_unused_variables,
)
from idaes.core.util.scaling import (
    get_jacobian,
    jacobian_cond,
)
from idaes.core.util.testing import (
    PhysicalParameterTestBlock,
    ReactionParameterTestBlock,
    initialization_tester,
)
from idaes.core.solvers import get_solver
from idaes.core.initialization import (
    BlockTriangularizationInitializer,
    SingleControlVolumeUnitInitializer,
    InitializationStatus,
)
from idaes.core.util import DiagnosticsToolbox
from idaes.core.scaling import set_scaling_factor

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver("ipopt_v2")


# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = PhysicalParameterTestBlock()
    m.fs.reactions = ReactionParameterTestBlock(property_package=m.fs.properties)

    m.fs.unit = EquilibriumReactor(
        property_package=m.fs.properties, reaction_package=m.fs.reactions
    )

    # Check unit config arguments
    assert len(m.fs.unit.config) == 15

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.material_balance_type == MaterialBalanceType.useDefault
    assert m.fs.unit.config.energy_balance_type == EnergyBalanceType.useDefault
    assert m.fs.unit.config.momentum_balance_type == MomentumBalanceType.pressureTotal
    assert m.fs.unit.config.has_rate_reactions
    assert not m.fs.unit.config.has_heat_transfer
    assert not m.fs.unit.config.has_pressure_change
    assert m.fs.unit.config.has_equilibrium_reactions
    assert not m.fs.unit.config.has_phase_equilibrium
    assert not m.fs.unit.config.has_heat_of_reaction
    assert m.fs.unit.config.property_package is m.fs.properties
    assert m.fs.unit.config.reaction_package is m.fs.reactions

    assert m.fs.unit.default_initializer is SingleControlVolumeUnitInitializer
    assert m.fs.unit.default_scaler is EquilibriumReactorScaler


# -----------------------------------------------------------------------------
class TestSaponification(object):
    @pytest.fixture(scope="class")
    def sapon(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = SaponificationParameterBlock()
        m.fs.reactions = SaponificationReactionParameterBlock(
            property_package=m.fs.properties
        )

        m.fs.unit = EquilibriumReactor(
            property_package=m.fs.properties,
            reaction_package=m.fs.reactions,
            has_equilibrium_reactions=False,
            has_heat_transfer=True,
            has_heat_of_reaction=True,
            has_pressure_change=True,
        )

        m.fs.unit.inlet.flow_vol.fix(1.0e-03)
        m.fs.unit.inlet.conc_mol_comp[0, "H2O"].fix(55388.0)
        m.fs.unit.inlet.conc_mol_comp[0, "NaOH"].fix(100.0)
        m.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        m.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].fix(1e-8)
        m.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].fix(1e-8)

        m.fs.unit.inlet.temperature.fix(303.15)
        m.fs.unit.inlet.pressure.fix(101325.0)

        m.fs.unit.heat_duty.fix(0)
        m.fs.unit.deltaP.fix(0)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, sapon):
        assert hasattr(sapon.fs.unit, "inlet")
        assert len(sapon.fs.unit.inlet.vars) == 4
        assert hasattr(sapon.fs.unit.inlet, "flow_vol")
        assert hasattr(sapon.fs.unit.inlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.inlet, "temperature")
        assert hasattr(sapon.fs.unit.inlet, "pressure")

        assert hasattr(sapon.fs.unit, "outlet")
        assert len(sapon.fs.unit.outlet.vars) == 4
        assert hasattr(sapon.fs.unit.outlet, "flow_vol")
        assert hasattr(sapon.fs.unit.outlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.outlet, "temperature")
        assert hasattr(sapon.fs.unit.outlet, "pressure")

        assert hasattr(sapon.fs.unit, "rate_reaction_constraint")
        assert hasattr(sapon.fs.unit, "heat_duty")
        assert hasattr(sapon.fs.unit, "deltaP")

        assert number_variables(sapon) == 26
        assert number_total_constraints(sapon) == 16
        assert number_unused_variables(sapon) == 0

    @pytest.mark.component
    def test_structural_issues(self, sapon):
        dt = DiagnosticsToolbox(sapon)
        dt.assert_no_structural_warnings()

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, sapon):
        initialization_tester(sapon)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, sapon):
        results = solver.solve(sapon)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, sapon):
        assert pytest.approx(101325.0, abs=1e-2) == value(
            sapon.fs.unit.outlet.pressure[0]
        )
        assert pytest.approx(304.32, abs=1e-2) == value(
            sapon.fs.unit.outlet.temperature[0]
        )
        assert pytest.approx(0.01, abs=1e-2) == value(
            sapon.fs.unit.outlet.conc_mol_comp[0, "EthylAcetate"]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, sapon):
        assert (
            abs(
                value(
                    sapon.fs.unit.inlet.flow_vol[0] - sapon.fs.unit.outlet.flow_vol[0]
                )
            )
            <= 1e-6
        )
        assert (
            abs(
                value(
                    sapon.fs.unit.inlet.flow_vol[0]
                    * sum(
                        sapon.fs.unit.inlet.conc_mol_comp[0, j]
                        for j in sapon.fs.properties.component_list
                    )
                    - sapon.fs.unit.outlet.flow_vol[0]
                    * sum(
                        sapon.fs.unit.outlet.conc_mol_comp[0, j]
                        for j in sapon.fs.properties.component_list
                    )
                )
            )
            <= 1e-6
        )
        assert (
            abs(
                value(
                    sapon.fs.unit.inlet.flow_vol[0]
                    * sapon.fs.properties.cp_mol
                    * (
                        sapon.fs.unit.inlet.temperature[0]
                        - sapon.fs.properties.temperature_ref
                    )
                    - sapon.fs.unit.outlet.flow_vol[0]
                    * sapon.fs.properties.cp_mol
                    * (
                        sapon.fs.unit.outlet.temperature[0]
                        - sapon.fs.properties.temperature_ref
                    )
                    + sapon.fs.unit.heat_duty[0]
                )
            )
            <= 1e-1
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_numerical_issues(self, sapon):
        dt = DiagnosticsToolbox(sapon)
        dt.assert_no_numerical_warnings()

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_performance_contents(self, sapon):
        perf_dict = sapon.fs.unit._get_performance_contents()

        assert perf_dict == {
            "vars": {
                "Heat Duty": sapon.fs.unit.heat_duty[0],
                "Pressure Change": sapon.fs.unit.deltaP[0],
            }
        }


class TestInitializers:
    @pytest.fixture
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = SaponificationParameterBlock()
        m.fs.reactions = SaponificationReactionParameterBlock(
            property_package=m.fs.properties
        )

        m.fs.unit = EquilibriumReactor(
            property_package=m.fs.properties,
            reaction_package=m.fs.reactions,
            has_equilibrium_reactions=False,
            has_heat_transfer=True,
            has_heat_of_reaction=True,
            has_pressure_change=True,
        )

        m.fs.unit.inlet.flow_vol[0].set_value(1.0e-03)
        m.fs.unit.inlet.conc_mol_comp[0, "H2O"].set_value(55388.0)
        m.fs.unit.inlet.conc_mol_comp[0, "NaOH"].set_value(100.0)
        m.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].set_value(100.0)
        m.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].set_value(0.0)
        m.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].set_value(0.0)

        m.fs.unit.inlet.temperature[0].set_value(303.15)
        m.fs.unit.inlet.pressure[0].set_value(101325.0)

        m.fs.unit.heat_duty.fix(0)
        m.fs.unit.deltaP.fix(0)

        return m

    @pytest.mark.component
    def test_general_hierarchical(self, model):
        initializer = SingleControlVolumeUnitInitializer()
        initializer.initialize(model.fs.unit)

        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

        assert value(model.fs.unit.outlet.flow_vol[0]) == pytest.approx(1e-3, rel=1e-6)
        assert value(model.fs.unit.outlet.conc_mol_comp[0, "H2O"]) == pytest.approx(
            55388, rel=1e-5
        )
        assert value(model.fs.unit.outlet.conc_mol_comp[0, "NaOH"]) == pytest.approx(
            0.01611, abs=1e-5
        )
        assert value(
            model.fs.unit.outlet.conc_mol_comp[0, "EthylAcetate"]
        ) == pytest.approx(0.01611, abs=1e-5)
        assert value(
            model.fs.unit.outlet.conc_mol_comp[0, "SodiumAcetate"]
        ) == pytest.approx(99.984, rel=1e-5)
        assert value(model.fs.unit.outlet.conc_mol_comp[0, "Ethanol"]) == pytest.approx(
            99.984, rel=1e-5
        )
        assert value(model.fs.unit.outlet.temperature[0]) == pytest.approx(
            304.32, rel=1e-2
        )
        assert value(model.fs.unit.outlet.pressure[0]) == pytest.approx(
            101325, rel=1e-6
        )

        assert not model.fs.unit.inlet.flow_vol[0].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "H2O"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "NaOH"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].fixed

        assert not model.fs.unit.inlet.temperature[0].fixed
        assert not model.fs.unit.inlet.pressure[0].fixed

    @pytest.mark.component
    def test_block_triangularization(self, model):
        initializer = BlockTriangularizationInitializer(
            constraint_tolerance=2e-5,
            block_solver_writer_config={"linear_presolve": False},
        )
        initializer.initialize(model.fs.unit)

        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

        assert value(model.fs.unit.outlet.flow_vol[0]) == pytest.approx(1e-3, rel=1e-6)
        assert value(model.fs.unit.outlet.conc_mol_comp[0, "H2O"]) == pytest.approx(
            55388, rel=1e-5
        )
        assert value(model.fs.unit.outlet.conc_mol_comp[0, "NaOH"]) == pytest.approx(
            0.0026102, rel=2e-5
        )
        assert value(
            model.fs.unit.outlet.conc_mol_comp[0, "EthylAcetate"]
        ) == pytest.approx(0.0026102, rel=2e-5)
        assert value(
            model.fs.unit.outlet.conc_mol_comp[0, "SodiumAcetate"]
        ) == pytest.approx(99.997, rel=2e-5)
        assert value(model.fs.unit.outlet.conc_mol_comp[0, "Ethanol"]) == pytest.approx(
            99.997, rel=2e-5
        )
        assert value(model.fs.unit.outlet.temperature[0]) == pytest.approx(
            304.32, rel=1e-2
        )
        assert value(model.fs.unit.outlet.pressure[0]) == pytest.approx(
            101325, rel=1e-6
        )

        assert not model.fs.unit.inlet.flow_vol[0].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "H2O"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "NaOH"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].fixed

        assert not model.fs.unit.inlet.temperature[0].fixed
        assert not model.fs.unit.inlet.pressure[0].fixed


class DummyScaler:
    def variable_scaling_routine(self, model, **kwargs):
        model._dummy_var_scaler = True

    def constraint_scaling_routine(self, model, **kwargs):
        model._dummy_con_scaler = True


class TestEquilibriumReactorScaler:
    @pytest.fixture
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = SaponificationParameterBlock()
        m.fs.reactions = SaponificationReactionParameterBlock(
            property_package=m.fs.properties
        )

        m.fs.unit = EquilibriumReactor(
            property_package=m.fs.properties,
            reaction_package=m.fs.reactions,
            has_equilibrium_reactions=False,
            has_heat_transfer=True,
            has_heat_of_reaction=True,
            has_pressure_change=True,
        )

        m.fs.unit.inlet.flow_vol[0].set_value(1.0e-03)
        m.fs.unit.inlet.conc_mol_comp[0, "H2O"].set_value(55388.0)
        m.fs.unit.inlet.conc_mol_comp[0, "NaOH"].set_value(100.0)
        m.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].set_value(100.0)
        m.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].set_value(0.0)
        m.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].set_value(0.0)

        m.fs.unit.inlet.temperature[0].set_value(303.15)
        m.fs.unit.inlet.pressure[0].set_value(101325.0)

        m.fs.unit.heat_duty.fix(0)
        m.fs.unit.deltaP.fix(0)

        return m

    @pytest.mark.component
    def test_variable_scaling_routine(self, model):
        scaler = model.fs.unit.default_scaler()

        assert isinstance(scaler, EquilibriumReactorScaler)

        scaler.variable_scaling_routine(model.fs.unit)

        # Inlet state
        sfx_in = model.fs.unit.control_volume.properties_in[0].scaling_factor
        assert isinstance(sfx_in, Suffix)
        assert len(sfx_in) == 8
        assert sfx_in[
            model.fs.unit.control_volume.properties_in[0].flow_vol
        ] == pytest.approx(1e2, rel=1e-8)
        assert sfx_in[
            model.fs.unit.control_volume.properties_in[0].pressure
        ] == pytest.approx(1e-5, rel=1e-8)
        assert sfx_in[
            model.fs.unit.control_volume.properties_in[0].temperature
        ] == pytest.approx(1 / 310.65, rel=1e-8)
        for k, v in model.fs.unit.control_volume.properties_in[0].conc_mol_comp.items():
            if k == "H2O":
                assert sfx_in[v] == pytest.approx(1e-4, rel=1e-8)
            else:
                assert sfx_in[v] == pytest.approx(1e-2, rel=1e-8)

        # Outlet state - should be the same as the inlet
        sfx_out = model.fs.unit.control_volume.properties_out[0].scaling_factor
        assert isinstance(sfx_out, Suffix)
        assert len(sfx_out) == 8
        assert sfx_out[
            model.fs.unit.control_volume.properties_out[0].flow_vol
        ] == pytest.approx(1e2, rel=1e-8)
        assert sfx_out[
            model.fs.unit.control_volume.properties_out[0].pressure
        ] == pytest.approx(1e-5, rel=1e-8)
        assert sfx_out[
            model.fs.unit.control_volume.properties_out[0].temperature
        ] == pytest.approx(1 / 310.65, rel=1e-8)
        for k, v in model.fs.unit.control_volume.properties_out[
            0
        ].conc_mol_comp.items():
            if k == "H2O":
                assert sfx_out[v] == pytest.approx(1e-4, rel=1e-8)
            else:
                assert sfx_out[v] == pytest.approx(1e-2, rel=1e-8)

        # Reaction block
        sfx_rxn = model.fs.unit.control_volume.reactions[0].scaling_factor
        assert isinstance(sfx_rxn, Suffix)
        assert len(sfx_rxn) == 2
        assert sfx_rxn[
            model.fs.unit.control_volume.reactions[0].k_rxn
        ] == pytest.approx(
            1 / (3.132e6 * exp(-43000 / (8.31446262 * 310.65))), rel=1e-8
        )
        assert sfx_rxn[
            model.fs.unit.control_volume.reactions[0].reaction_rate["R1"]
        ] == pytest.approx(1e2, rel=1e-8)

        # Check that unit model has scaling factors
        sfx_cv = model.fs.unit.control_volume.scaling_factor
        assert isinstance(sfx_cv, Suffix)
        assert len(sfx_cv) == 2
        assert sfx_cv[model.fs.unit.control_volume.heat[0]] == pytest.approx(
            1.917448e-05, rel=1e-3
        )
        assert sfx_cv[model.fs.unit.control_volume.deltaP[0]] == pytest.approx(
            1e-4, rel=1e-3
        )

        # No unit level variables to scale, so no suffix
        assert not hasattr(model.fs.unit, "scaling_factor")

    @pytest.mark.component
    def test_variable_scaling_routine_submodel_scaler(self, model):
        scaler = model.fs.unit.default_scaler()

        scaler_map = ComponentMap()
        scaler_map[model.fs.unit.control_volume.properties_in] = DummyScaler
        scaler_map[model.fs.unit.control_volume.properties_out] = DummyScaler
        scaler_map[model.fs.unit.control_volume.reactions] = DummyScaler

        scaler.variable_scaling_routine(
            model.fs.unit,
            submodel_scalers=scaler_map,
        )

        # Should call DummyScaler submethod for each submodel
        # Should add _dummy_var_scaler = True to all submodels
        assert model.fs.unit.control_volume.properties_in[0]._dummy_var_scaler
        assert model.fs.unit.control_volume.properties_out[0]._dummy_var_scaler
        assert model.fs.unit.control_volume.reactions[0]._dummy_var_scaler

    @pytest.mark.component
    def test_constraint_scaling_routine(self, model):
        scaler = model.fs.unit.default_scaler()

        assert isinstance(scaler, EquilibriumReactorScaler)

        scaler.constraint_scaling_routine(model.fs.unit)

        # Check that sub-models have suffixes - we will assume they are right at this point
        sfx_in = model.fs.unit.control_volume.properties_in[0].scaling_factor
        assert isinstance(sfx_in, Suffix)
        assert (
            len(sfx_in) == 0
        )  # inlet has no constraints. Not quite sure why the Suffix exists

        sfx_out = model.fs.unit.control_volume.properties_out[0].scaling_factor
        assert isinstance(sfx_out, Suffix)
        assert len(sfx_out) == 1
        assert sfx_out[
            model.fs.unit.control_volume.properties_out[0.0].conc_water_eqn
        ] == pytest.approx(1e-4, rel=1e-8)

        sfx_rxn = model.fs.unit.control_volume.reactions[0].scaling_factor
        assert isinstance(sfx_rxn, Suffix)
        assert len(sfx_rxn) == 2
        assert sfx_rxn[
            model.fs.unit.control_volume.reactions[0.0].arrhenius_eqn
        ] == pytest.approx(1, rel=1e-8)
        assert sfx_rxn[
            model.fs.unit.control_volume.reactions[0.0].rate_expression["R1"]
        ] == pytest.approx(1e-4, rel=1e-8)

        # Check that unit model has scaling factors
        sfx_cv = model.fs.unit.control_volume.scaling_factor
        assert isinstance(model.fs.unit.control_volume.scaling_factor, Suffix)
        assert len(sfx_cv) == 12
        assert sfx_cv[
            model.fs.unit.control_volume.enthalpy_balances[0.0]
        ] == pytest.approx(8.03894082e-10, rel=1e-8)
        assert sfx_cv[
            model.fs.unit.control_volume.pressure_balance[0.0]
        ] == pytest.approx(9.86923267e-6, rel=1e-8)
        for c in model.fs.unit.control_volume.material_balances.values():
            assert sfx_cv[c] == pytest.approx(1e-2, rel=1e-8)
        for (
            c
        ) in (
            model.fs.unit.control_volume.rate_reaction_stoichiometry_constraint.values()
        ):
            assert sfx_cv[c] == pytest.approx(1, rel=1e-8)

        sfx_unit = model.fs.unit.scaling_factor
        assert isinstance(sfx_unit, Suffix)
        assert len(sfx_unit) == 1
        assert sfx_unit[
            model.fs.unit.rate_reaction_constraint[0.0, "R1"]
        ] == pytest.approx(1, rel=1e-8)

    @pytest.mark.component
    def test_constraint_scaling_routine_submodel_scaler(self, model):
        scaler = model.fs.unit.default_scaler()

        scaler_map = ComponentMap()
        scaler_map[model.fs.unit.control_volume.properties_in] = DummyScaler
        scaler_map[model.fs.unit.control_volume.properties_out] = DummyScaler
        scaler_map[model.fs.unit.control_volume.reactions] = DummyScaler

        scaler.constraint_scaling_routine(
            model.fs.unit,
            submodel_scalers=scaler_map,
        )

        # Should call DummyScaler submethod for each submodel
        # Should add _dummy_con_scaler = True to all submodels
        assert model.fs.unit.control_volume.properties_in[0]._dummy_con_scaler
        assert model.fs.unit.control_volume.properties_out[0]._dummy_con_scaler
        assert model.fs.unit.control_volume.reactions[0]._dummy_con_scaler

    @pytest.mark.component
    def test_scale_model(self, model):
        scaler = model.fs.unit.default_scaler()

        assert isinstance(scaler, EquilibriumReactorScaler)

        scaler.scale_model(model.fs.unit)

        # Inlet state
        sfx_in = model.fs.unit.control_volume.properties_in[0].scaling_factor
        assert isinstance(sfx_in, Suffix)
        assert len(sfx_in) == 8
        assert sfx_in[
            model.fs.unit.control_volume.properties_in[0].flow_vol
        ] == pytest.approx(1e2, rel=1e-8)
        assert sfx_in[
            model.fs.unit.control_volume.properties_in[0].pressure
        ] == pytest.approx(1e-5, rel=1e-8)
        assert sfx_in[
            model.fs.unit.control_volume.properties_in[0].temperature
        ] == pytest.approx(1 / 310.65, rel=1e-8)
        for k, v in model.fs.unit.control_volume.properties_in[0].conc_mol_comp.items():
            if k == "H2O":
                assert sfx_in[v] == pytest.approx(1e-4, rel=1e-8)
            else:
                assert sfx_in[v] == pytest.approx(1e-2, rel=1e-8)

        # Outlet state - should be the same as the inlet
        sfx_out = model.fs.unit.control_volume.properties_out[0].scaling_factor
        assert isinstance(sfx_out, Suffix)
        assert len(sfx_out) == 9
        assert sfx_out[
            model.fs.unit.control_volume.properties_out[0].flow_vol
        ] == pytest.approx(1e2, rel=1e-8)
        assert sfx_out[
            model.fs.unit.control_volume.properties_out[0].pressure
        ] == pytest.approx(1e-5, rel=1e-8)
        assert sfx_out[
            model.fs.unit.control_volume.properties_out[0].temperature
        ] == pytest.approx(1 / 310.65, rel=1e-8)
        for k, v in model.fs.unit.control_volume.properties_out[
            0
        ].conc_mol_comp.items():
            if k == "H2O":
                assert sfx_out[v] == pytest.approx(1e-4, rel=1e-8)
            else:
                assert sfx_out[v] == pytest.approx(1e-2, rel=1e-8)
        assert sfx_out[
            model.fs.unit.control_volume.properties_out[0.0].conc_water_eqn
        ] == pytest.approx(1e-4, rel=1e-8)

        # Reaction block
        sfx_rxn = model.fs.unit.control_volume.reactions[0].scaling_factor
        assert isinstance(sfx_rxn, Suffix)
        assert len(sfx_rxn) == 4
        assert sfx_rxn[
            model.fs.unit.control_volume.reactions[0].k_rxn
        ] == pytest.approx(
            1 / (3.132e6 * exp(-43000 / (8.31446262 * 310.65))), rel=1e-8
        )
        assert sfx_rxn[
            model.fs.unit.control_volume.reactions[0].reaction_rate["R1"]
        ] == pytest.approx(1e2, rel=1e-8)
        assert sfx_rxn[
            model.fs.unit.control_volume.reactions[0.0].arrhenius_eqn
        ] == pytest.approx(5.4240896, rel=1e-8)
        assert sfx_rxn[
            model.fs.unit.control_volume.reactions[0.0].rate_expression["R1"]
        ] == pytest.approx(5.4240896e-4, rel=1e-8)

        # Check that unit model has scaling factors
        sfx_cv = model.fs.unit.control_volume.scaling_factor
        assert isinstance(sfx_cv, Suffix)
        assert len(sfx_cv) == 14
        assert sfx_cv[model.fs.unit.control_volume.heat[0]] == pytest.approx(
            1.917448e-05, rel=1e-3
        )
        assert sfx_cv[model.fs.unit.control_volume.deltaP[0]] == pytest.approx(
            1e-4, rel=1e-3
        )
        assert sfx_cv[
            model.fs.unit.control_volume.enthalpy_balances[0.0]
        ] == pytest.approx(7.71546823e-08, rel=1e-8)
        assert sfx_cv[
            model.fs.unit.control_volume.pressure_balance[0.0]
        ] == pytest.approx(1e-5, rel=1e-8)
        for (_, _, j), c in model.fs.unit.control_volume.material_balances.items():
            if j == "H2O":
                assert sfx_cv[c] == pytest.approx(1e-2, rel=1e-8)
            else:
                assert sfx_cv[c] == pytest.approx(1, rel=1e-8)
        for (
            _,
            _,
            j,
        ), c in (
            model.fs.unit.control_volume.rate_reaction_stoichiometry_constraint.items()
        ):
            assert sfx_cv[c] == pytest.approx(1, rel=1e-8)

        sfx_unit = model.fs.unit.scaling_factor
        assert isinstance(sfx_unit, Suffix)
        assert len(sfx_unit) == 1
        assert sfx_unit[
            model.fs.unit.rate_reaction_constraint[0.0, "R1"]
        ] == pytest.approx(1e2, rel=1e-8)

    @pytest.mark.integration
    def test_example_case(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = SaponificationParameterBlock()
        m.fs.reactions = SaponificationReactionParameterBlock(
            property_package=m.fs.properties
        )

        m.fs.equil = EquilibriumReactor(
            property_package=m.fs.properties,
            reaction_package=m.fs.reactions,
            has_equilibrium_reactions=False,
            has_heat_of_reaction=True,
        )

        m.fs.equil.inlet.flow_vol.fix(1.0e-03)
        m.fs.equil.inlet.conc_mol_comp[0, "H2O"].fix(55388.0)
        m.fs.equil.inlet.conc_mol_comp[0, "NaOH"].fix(100.0)
        m.fs.equil.inlet.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        m.fs.equil.inlet.conc_mol_comp[0, "SodiumAcetate"].fix(1e-8)
        m.fs.equil.inlet.conc_mol_comp[0, "Ethanol"].fix(1e-8)

        m.fs.equil.inlet.temperature.fix(303.15)
        m.fs.equil.inlet.pressure.fix(101325.0)

        initializer = BlockTriangularizationInitializer()
        initializer.initialize(m.fs.equil)

        set_scaling_factor(m.fs.equil.control_volume.properties_in[0].flow_vol, 1e3)

        scaler = EquilibriumReactorScaler()
        scaler.scale_model(m.fs.equil)

        m.fs.equil.inlet.flow_vol.fix(1)
        m.fs.equil.inlet.conc_mol_comp[0, "NaOH"].fix(200.0)
        m.fs.equil.inlet.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        m.fs.equil.inlet.conc_mol_comp[0, "SodiumAcetate"].fix(50)
        m.fs.equil.inlet.conc_mol_comp[0, "Ethanol"].fix(1e-8)

        m.fs.equil.inlet.temperature.fix(320)

        solver = get_solver(
            "ipopt_v2", writer_config={"linear_presolve": True, "scale_model": True}
        )
        results = solver.solve(m, tee=True)
        assert_optimal_termination(results)

        # Check condition number to confirm scaling
        sm = TransformationFactory("core.scale_model").create_using(m, rename=False)
        jac, _ = get_jacobian(sm, scaled=False)
        assert (jacobian_cond(jac=jac, scaled=False)) == pytest.approx(
            4.987e05, rel=1e-3
        )



# -*- coding: utf-8 -*-
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
""" Variable classes for time-indexed variables that can be treated specially
by NMPC.
"""

from pyomo.core.base.var import IndexedVar
from pyomo.core.base.indexed_component_slice import IndexedComponent_slice


class NmpcVar(IndexedVar):
    """
    This class is meant to be used as a ctype for references to time-only
    slices. It provides attributes that we would like to store for each
    "variable," but not for different time-indices of the same "variable."
    Consistent with the NMPC/DAE literature, a "variable" here is
    something indexed only by time.

    Users may want to access these "time-only variables," and may do so
    by looping over `component_objects`. For example:

    >>> for var in model.component_objects(NmpcVar):
    >>>     var[:].set_value(var[0])

    """

    def __init__(self, *args, **kwargs):
        if not args:
            raise NotImplementedError(
                "%s component must be indexed by at least one set." % self.__class__
            )
        self.setpoint = kwargs.pop("setpoint", None)
        self.weight = kwargs.pop("weight", None)
        self.variance = kwargs.pop("variance", None)
        self.nominal = kwargs.pop("nominal", None)
        self.noise_bounds = kwargs.pop("noise_bounds", None)
        kwargs.setdefault("ctype", type(self))
        super(NmpcVar, self).__init__(*args, **kwargs)


"""
The following classes serve as custom ctypes for references
to time-only slices and calls to `component_objects` when
the user or developer wants to specify a particular subset
of the "NMPC variables."
"""


class DiffVar(NmpcVar):
    _attr = "differential"


class DerivVar(NmpcVar):
    _attr = "derivative"


class AlgVar(NmpcVar):
    _attr = "algebraic"


class InputVar(NmpcVar):
    _attr = "input"


class FixedVar(NmpcVar):
    _attr = "fixed"


class MeasuredVar(NmpcVar):
    _attr = "measurement"


class _NmpcVector(IndexedVar):
    """
    This is a class used as a ctype for references to `NmpcVar` objects.
    It is used to construct a "vector" of these variables with a
    "Pyomothonic" API. These "vectors" are indexed by an integer describing
    the "coordinate" of the variable within the vector and by time.
    In addition to accessing the underlying data objects with a 2-dimensional
    index, this class adds methods to iterate over the underlying `NmpcVar`
    objects. This is necessary to access the special attributes on these
    variables that would not make sense to attach to a data object.

    For example:

    >>> input_vars = Reference(INPUT_BLOCK[:].var[:], ctype=_NmpcVector)
    >>> input_vars[:, 0].fix()
    >>> input_vars.set_setpoint(input_setpoint_list)

    """

    def _generate_referenced_vars(self):
        _slice = self.referent.duplicate()
        # This class should only be instantiated as a reference...
        assert type(_slice) is IndexedComponent_slice
        # _slice looks something like BLOCK[:].var[:]
        assert _slice._call_stack[-1][0] == IndexedComponent_slice.get_item
        assert _slice._call_stack[-1][1] == slice(None)
        assert _slice._call_stack[-2][0] == IndexedComponent_slice.get_attribute
        assert _slice._call_stack[-3][0] == IndexedComponent_slice.slice_info
        _slice._call_stack.pop()
        _slice._len -= 1
        for var in _slice:
            assert isinstance(var, NmpcVar)
            yield var

    def set_setpoint(self, setpoint):
        referent_gen = self._generate_referenced_vars()
        try:
            for var, sp in zip(referent_gen, setpoint):
                var.setpoint = sp
        except TypeError:
            for var in self._generate_referenced_vars():
                var.setpoint = setpoint

    def get_setpoint(self):
        for var in self._generate_referenced_vars():
            yield var.setpoint

    @property
    def values(self):
        referent_gen = self._generate_referenced_vars()
        return list(list(var[t].value for t in var) for var in referent_gen)

    @values.setter
    def values(self, vals):
        referent_gen = self._generate_referenced_vars()
        try:
            for var, val in zip(referent_gen, vals):
                # var is time-indexed
                var[:].set_value(val)
        except TypeError:
            # A scalar value was provided. Set for all i, t
            self[...].set_value(vals)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for Phase objects

Author: Andrew Lee
"""
from pyomo.environ import ConcreteModel, Set

from idaes.core.base.phases import (
    Phase,
    LiquidPhase,
    SolidPhase,
    VaporPhase,
    PhaseType,
    AqueousPhase,
)
import pytest


@pytest.mark.unit
def test_PhaseType():
    assert len(PhaseType) == 5
    for i in PhaseType.__members__:
        assert i in [
            "undefined",
            "liquidPhase",
            "vaporPhase",
            "solidPhase",
            "aqueousPhase",
        ]


@pytest.mark.unit
def test_config():
    m = ConcreteModel()

    m.phase = Phase()

    for k, v in m.phase.config.items():
        if k == "_phase_list_exists":
            assert not v
        elif k == "parameter_data" or k == "transport_property_options":
            assert v == {}
        else:
            assert v is None


@pytest.mark.unit
def test_populate_phase_list():
    m = ConcreteModel()

    m.phase = Phase()
    m.phase2 = Phase()

    assert isinstance(m.phase_list, Set)

    for p in m.phase_list:
        assert p in ["phase", "phase2"]


@pytest.mark.unit
def test_is_phase_generic():
    m = ConcreteModel()

    m.phase = Phase()

    assert not m.phase.is_liquid_phase()
    assert not m.phase.is_solid_phase()
    assert not m.phase.is_vapor_phase()
    assert not m.phase.is_aqueous_phase()


@pytest.mark.unit
def test_is_phase_old_style_liquid():
    m = ConcreteModel()

    m.Liq = Phase()

    assert m.Liq.is_liquid_phase()
    assert not m.Liq.is_solid_phase()
    assert not m.Liq.is_vapor_phase()
    assert not m.Liq.is_aqueous_phase()


@pytest.mark.unit
def test_is_phase_old_style_solid():
    m = ConcreteModel()

    m.Sol = Phase()

    assert not m.Sol.is_liquid_phase()
    assert m.Sol.is_solid_phase()
    assert not m.Sol.is_vapor_phase()
    assert not m.Sol.is_aqueous_phase()


@pytest.mark.unit
def test_is_phase_old_style_vapor():
    m = ConcreteModel()

    m.Vap = Phase()

    assert not m.Vap.is_liquid_phase()
    assert not m.Vap.is_solid_phase()
    assert m.Vap.is_vapor_phase()
    assert not m.Vap.is_aqueous_phase()


@pytest.mark.unit
def test_phase_list_exists():
    m = ConcreteModel()

    m.phase = Phase(_phase_list_exists=True)
    m.phase2 = Phase(_phase_list_exists=True)

    assert not hasattr(m, "phase_list")


@pytest.mark.unit
def test_LiquidPhase():
    m = ConcreteModel()

    m.phase = LiquidPhase()

    assert m.phase.is_liquid_phase()
    assert not m.phase.is_solid_phase()
    assert not m.phase.is_vapor_phase()
    assert not m.phase.is_aqueous_phase()


@pytest.mark.unit
def test_SolidPhase():
    m = ConcreteModel()

    m.phase = SolidPhase()

    assert not m.phase.is_liquid_phase()
    assert m.phase.is_solid_phase()
    assert not m.phase.is_vapor_phase()
    assert not m.phase.is_aqueous_phase()


@pytest.mark.unit
def test_VaporPhase():
    m = ConcreteModel()

    m.phase = VaporPhase()

    assert not m.phase.is_liquid_phase()
    assert not m.phase.is_solid_phase()
    assert m.phase.is_vapor_phase()
    assert not m.phase.is_aqueous_phase()


@pytest.mark.unit
def test_AqueousPhase():
    m = ConcreteModel()

    m.phase = AqueousPhase()

    assert m.phase.is_liquid_phase()
    assert not m.phase.is_solid_phase()
    assert not m.phase.is_vapor_phase()
    assert m.phase.is_aqueous_phase()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Dictionary of surrogate models

@author: Boiler Subsystem Team (J. Ma, M. Zamarripa)
"""
# A dictionary of surrogate models needs to be provided for the boiler fire
# side model to work. Note that the user needs to provide a surrogate model or
# a fixed value for the heat flux for each water wall section, platen super
# heater (if installed), roof (if installed), flyash, and NOx prediction.
# in this test case the model builds 12 water wall zones, a platen sh, and roof
# usually a surrogate model is a function of the input variables:
# b.wall_temperature_waterwall[t, 1]
# b.wall_temperature_waterwall[t, 2]
# b.wall_temperature_waterwall[t, 3]
# b.wall_temperature_waterwall[t, 4]
# b.wall_temperature_waterwall[t, 5]
# b.wall_temperature_waterwall[t, 6]
# b.wall_temperature_waterwall[t, 7]
# b.wall_temperature_waterwall[t, 8]
# b.wall_temperature_waterwall[t, 9]
# b.wall_temperature_waterwall[t, 10] - or as many water wall zones
# b.wall_temperature_platen[t] - if has_platen_superheater=True
# b.wall_temperature_roof[t] - if has_roof_superheater=True
# b.flowrate_coal_raw[t]
# b.mf_H2O_coal_raw[t] - moisture content of raw coal
# b.SR[t] - stoichiometric ratio
# b.SR_lf[t] - lower furnace stoichiometric ratio
# b.secondary_air_inlet.temperature[t]
# b.ratio_PA2coal[t] - ratio primary air to coal flowrate calculated by model

# Since the development of surrogate models requires a specific case and
# the fire side must be connected with other unit models. This example aims
# testing the model. Therefore, the following dictionary is a simple example
# using fixed values for heat flux to the water wall, platen SH, and roof SH.
data_dic = {
    1: "2.0e7",  # heat flux to water wall zone 1 from fire side in W
    2: "1.0e7",  # heat flux to water wall zone 2 from fire side in W
    3: "1.0e7",  # heat flux to water wall zone 3 from fire side in W
    4: "1.0e7",  # heat flux to water wall zone 4 from fire side in W
    5: "1.5e7",  # heat flux to water wall zone 5 from fire side in W
    6: "1.0e7",  # heat flux to water wall zone 6 from fire side in W
    7: "1.2e7",  # heat flux to water wall zone 7 from fire side in W
    8: "3e7",  # heat flux to water wall zone 8 from fire side in W
    9: "2.5e7",  # heat flux to water wall zone 9 from fire side in W
    10: "2.0e7",  # heat flux to water wall zone 10 from fire side in W
    11: "1.8e7",  # heat flux to water wall zone 11 from fire side in W
    12: "1.0e7",  # heat flux to water wall zone 12 from fire side in W
    "pl": "5.0e7",  # heat flux to platen sh from fire side in W
    "roof": "6.5e7",  # heat flux to roof from fire side in W
    "flyash": "0.0001",  # flyash or unburned carbon mass fraction
    "NOx": "140",
}  # NOx PPM



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
IDAES plate heat exchanger model (PHE) using effectiveness-NTU method, derived
from the core HeatExchangerNTU model.

Notes:
Let P = number of passes(same for both fluids)
if P is even PHE is in counter-current mode
if P is odd PHE  is in co-current mode
Hot and cold fluids flow alternatively in the channels.
Divider plates may be used to partition the passes in separate sections.
The heat transfer area  of a single plate depends on the total heat transfer
area as specified by the manufacturer and the total number of active plates.

Reference
Detailed model equations can be found in the paper :
[1] Akula, P., Eslick, J., Bhattacharyya, D. and Miller, D.C., 2019.
Modelling and Parameter Estimation of a Plate Heat Exchanger
as Part of a Solvent-Based Post-Combustion CO2 Capture System.
In Computer Aided Chemical Engineering (Vol. 47, pp. 47-52). Elsevier.

"""

# Import Pyomo libraries
from pyomo.environ import (
    Constraint,
    exp,
    Expression,
    NonNegativeIntegers,
    Param,
    PositiveIntegers,
    PositiveReals,
    units as pyunits,
    Var,
)
from pyomo.common.config import ConfigValue, In, Integer
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from pyomo.common.deprecation import deprecated

# Import IDAES cores
from idaes.core import declare_process_block_class
from idaes.models.unit_models.heat_exchanger_ntu import HeatExchangerNTUData
from idaes.core.util.constants import Constants
from idaes.core.solvers import get_solver
import idaes.logger as idaeslog

__author__ = "Paul Akula, Andrew Lee"


# Set up logger
_log = idaeslog.getLogger(__name__)


@deprecated(
    "The Plate Heat Exchanger (PHE) model is known to be affected by "
    "issues causing it to fail to solve on certain platforms starting with Pyomo v6.7.0. "
    "This might cause the model to be removed in an upcoming IDAES release if these failures are not resolved. "
    "For more information, see IDAES/idaes-pse#1294",
    version="2.3.0",
)
@declare_process_block_class("PlateHeatExchanger")
class PlateHeatExchangerData(HeatExchangerNTUData):
    """Plate Heat Exchanger(PHE) Unit Model."""

    CONFIG = HeatExchangerNTUData.CONFIG()

    CONFIG.declare(
        "passes",
        ConfigValue(
            default=4,
            domain=Integer,
            description="Number of passes",
            doc="""Number of passes of the fluids through the heat exchanger""",
        ),
    )
    CONFIG.declare(
        "channels_per_pass",
        ConfigValue(
            default=12,
            domain=Integer,
            description="Number of channels for each pass",
            doc="""Number of channels to be used in each pass where a channel
               is the space between two plates with a flowing fluid""",
        ),
    )
    CONFIG.declare(
        "number_of_divider_plates",
        ConfigValue(
            default=0,
            domain=Integer,
            description="Number of divider plates in heat exchanger",
            doc="""Divider plates are used to create separate partitions in the
        unit. Each pass can be separated by a divider plate""",
        ),
    )

    # Update config block setting for pressure change to always be true
    CONFIG.hot_side.has_pressure_change = True
    CONFIG.hot_side.get("has_pressure_change").set_domain(In([True]))
    # pylint: disable-next=protected-access
    CONFIG.hot_side.get("has_pressure_change")._description = (
        "Pressure change term construction flag - must be True"
    )
    # pylint: disable-next=protected-access
    CONFIG.hot_side.get("has_pressure_change")._doc = (
        "Plate Heat Exchanger model includes correlations for pressure drop "
        "thus has_pressure_change must be True"
    )

    CONFIG.cold_side.has_pressure_change = True
    CONFIG.cold_side.get("has_pressure_change").set_domain(In([True]))
    # pylint: disable-next=protected-access
    CONFIG.cold_side.get("has_pressure_change")._description = (
        "Pressure change term construction flag - must be True"
    )
    # pylint: disable-next=protected-access
    CONFIG.cold_side.get("has_pressure_change")._doc = (
        "Plate Heat Exchanger model includes correlations for pressure drop "
        "thus has_pressure_change must be True"
    )

    def build(self):
        # Call super.build to setup model
        # This will create the control volumes, ports and basic equations
        super().build()

        # Units will be based on hot side properties
        units_meta = (
            self.hot_side.config.property_package.get_metadata().get_derived_units
        )

        # ---------------------------------------------------------------------
        # Plate design variables and parameter
        self.number_of_passes = Param(
            initialize=self.config.passes,
            units=pyunits.dimensionless,
            domain=PositiveIntegers,
            doc="Number of hot/cold fluid passes",
            mutable=True,
        )

        # Assuming number of channels is equal in all plates
        self.channels_per_pass = Param(
            initialize=self.config.channels_per_pass,
            units=pyunits.dimensionless,
            domain=PositiveIntegers,
            doc="Number of channels in each pass",
            mutable=True,
        )

        self.number_of_divider_plates = Param(
            initialize=self.config.number_of_divider_plates,
            units=pyunits.dimensionless,
            domain=NonNegativeIntegers,
            doc="Number of divider plates in heat exchanger",
            mutable=True,
        )

        self.plate_length = Var(
            initialize=1.6925,
            units=units_meta("length"),
            domain=PositiveReals,
            doc="Length of heat exchanger plate",
        )
        self.plate_width = Var(
            initialize=0.6135,
            units=units_meta("length"),
            domain=PositiveReals,
            doc="Width of heat exchanger plate",
        )
        self.plate_thickness = Var(
            initialize=0.0006,
            units=units_meta("length"),
            domain=PositiveReals,
            doc="Thickness of heat exchanger plate",
        )
        self.plate_pact_length = Var(
            initialize=0.381,
            units=units_meta("length"),
            domain=PositiveReals,
            doc="Compressed plate pact length",
        )
        self.port_diameter = Var(
            initialize=0.2045,
            units=units_meta("length"),
            domain=PositiveReals,
            doc="Port diameter",
        )

        self.plate_therm_cond = Var(
            initialize=16.2,
            units=units_meta("thermal_conductivity"),
            domain=PositiveReals,
            doc="Thermal conductivity heat exchanger plates",
        )

        # Set default value of total heat transfer area
        self.area.set_value(114.3)

        # ---------------------------------------------------------------------
        # Derived geometric quantities
        total_plates = (
            2 * self.channels_per_pass * self.number_of_passes
            + 1
            + self.number_of_divider_plates
        )
        total_active_plates = 2 * self.channels_per_pass * self.number_of_passes - (
            1 + self.number_of_divider_plates
        )

        self.plate_gap = Expression(
            expr=self.plate_pact_length / total_plates - self.plate_thickness
        )

        self.plate_area = Expression(
            expr=self.area / total_active_plates,
            doc="Heat transfer area of single plate",
        )

        self.surface_enlargement_factor = Expression(
            expr=self.plate_area / (self.plate_length * self.plate_width)
        )

        # Channel equivalent diameter
        self.channel_diameter = Expression(
            expr=2 * self.plate_gap / self.surface_enlargement_factor,
            doc="Channel equivalent diameter",
        )

        # ---------------------------------------------------------------------
        # Fluid velocities
        def rule_port_vel_hot(blk, t):
            return (
                4
                * blk.hot_side.properties_in[t].flow_vol
                / (Constants.pi * blk.port_diameter**2)
            )

        self.hot_port_velocity = Expression(
            self.flowsheet().time, rule=rule_port_vel_hot, doc="Hot side port velocity"
        )

        def rule_port_vel_cold(blk, t):
            return (
                4
                * pyunits.convert(
                    blk.cold_side.properties_in[t].flow_vol,
                    to_units=units_meta("flow_vol"),
                )
                / (Constants.pi * blk.port_diameter**2)
            )

        self.cold_port_velocity = Expression(
            self.flowsheet().time,
            rule=rule_port_vel_cold,
            doc="Cold side port velocity",
        )

        def rule_channel_vel_hot(blk, t):
            return blk.hot_side.properties_in[t].flow_vol / (
                blk.channels_per_pass * blk.plate_width * blk.plate_gap
            )

        self.hot_channel_velocity = Expression(
            self.flowsheet().time,
            rule=rule_channel_vel_hot,
            doc="Hot side channel velocity",
        )

        def rule_channel_vel_cold(blk, t):
            return pyunits.convert(
                blk.cold_side.properties_in[t].flow_vol, to_units=units_meta("flow_vol")
            ) / (blk.channels_per_pass * blk.plate_width * blk.plate_gap)

        self.cold_channel_velocity = Expression(
            self.flowsheet().time,
            rule=rule_channel_vel_cold,
            doc="Cold side channel velocity",
        )

        # ---------------------------------------------------------------------
        # Reynolds & Prandtl numbers
        # Density cancels out of Reynolds number if mass flow rate is used
        def rule_Re_h(blk, t):
            return (
                blk.hot_side.properties_in[t].flow_mass
                * blk.channel_diameter
                / (
                    blk.channels_per_pass
                    * blk.plate_width
                    * blk.plate_gap
                    * blk.hot_side.properties_in[t].visc_d_phase["Liq"]
                )
            )

        self.Re_hot = Expression(
            self.flowsheet().time, rule=rule_Re_h, doc="Hot side Reynolds number"
        )

        def rule_Re_c(blk, t):
            return (
                pyunits.convert(
                    blk.cold_side.properties_in[t].flow_mass
                    / blk.cold_side.properties_in[t].visc_d_phase["Liq"],
                    to_units=units_meta("length"),
                )
                * blk.channel_diameter
                / (blk.channels_per_pass * blk.plate_width * blk.plate_gap)
            )

        self.Re_cold = Expression(
            self.flowsheet().time, rule=rule_Re_c, doc="Cold side Reynolds number"
        )

        def rule_Pr_h(blk, t):
            return (
                blk.hot_side.properties_in[t].cp_mol
                / blk.hot_side.properties_in[t].mw
                * blk.hot_side.properties_in[t].visc_d_phase["Liq"]
                / blk.hot_side.properties_in[t].therm_cond_phase["Liq"]
            )

        self.Pr_hot = Expression(
            self.flowsheet().time, rule=rule_Pr_h, doc="Hot side Prandtl number"
        )

        def rule_Pr_c(blk, t):
            return (
                blk.cold_side.properties_in[t].cp_mol
                / blk.cold_side.properties_in[t].mw
                * blk.cold_side.properties_in[t].visc_d_phase["Liq"]
                / blk.cold_side.properties_in[t].therm_cond_phase["Liq"]
            )

        self.Pr_cold = Expression(
            self.flowsheet().time, rule=rule_Pr_c, doc="Cold side Prandtl number"
        )

        # ---------------------------------------------------------------------
        # Heat transfer coefficients
        # Parameters for Nusselt number correlation
        self.Nusselt_param_a = Param(
            initialize=0.4,
            domain=PositiveReals,
            units=pyunits.dimensionless,
            mutable=True,
            doc="Nusselt parameter A",
        )
        self.Nusselt_param_b = Param(
            initialize=0.663,
            domain=PositiveReals,
            units=pyunits.dimensionless,
            mutable=True,
            doc="Nusselt parameter B",
        )
        self.Nusselt_param_c = Param(
            initialize=0.333,
            domain=PositiveReals,
            units=pyunits.dimensionless,
            mutable=True,
            doc="Nusselt parameter C",
        )

        # Film heat transfer coefficients
        def rule_hotside_transfer_coeff(blk, t):
            return (
                blk.hot_side.properties_in[t].therm_cond_phase["Liq"]
                / blk.channel_diameter
                * blk.Nusselt_param_a
                * blk.Re_hot[t] ** blk.Nusselt_param_b
                * blk.Pr_hot[t] ** blk.Nusselt_param_c
            )

        self.heat_transfer_coefficient_hot_side = Expression(
            self.flowsheet().time,
            rule=rule_hotside_transfer_coeff,
            doc="Hot side heat transfer coefficient",
        )

        def rule_coldside_transfer_coeff(blk, t):
            return (
                pyunits.convert(
                    blk.cold_side.properties_in[t].therm_cond_phase["Liq"],
                    to_units=units_meta("thermal_conductivity"),
                )
                / blk.channel_diameter
                * blk.Nusselt_param_a
                * blk.Re_cold[t] ** blk.Nusselt_param_b
                * blk.Pr_cold[t] ** blk.Nusselt_param_c
            )

        self.heat_transfer_coefficient_cold_side = Expression(
            self.flowsheet().time,
            rule=rule_coldside_transfer_coeff,
            doc="Cold side heat transfer coefficient",
        )

        # Overall heat transfer coefficient
        def rule_U(blk, t):
            return blk.heat_transfer_coefficient[t] == (
                1.0
                / (
                    1.0 / blk.heat_transfer_coefficient_hot_side[t]
                    + blk.plate_gap / blk.plate_therm_cond
                    + 1.0 / blk.heat_transfer_coefficient_cold_side[t]
                )
            )

        self.overall_heat_transfer_eq = Constraint(
            self.flowsheet().time,
            rule=rule_U,
            doc="Calculations of overall heat transfer coefficient",
        )

        # Effectiveness based on sub-heat exchangers
        # Divide NTU by number of channels per pass
        def rule_Ecf(blk, t):
            if blk.number_of_passes.value % 2 == 0:
                return blk.effectiveness[t] == (
                    1 - exp(-blk.NTU[t] / blk.channels_per_pass * (1 - blk.Cratio[t]))
                ) / (
                    1
                    - blk.Cratio[t]
                    * exp(-blk.NTU[t] / blk.channels_per_pass * (1 - blk.Cratio[t]))
                )
            elif blk.pass_num.value % 2 == 1:
                return blk.effectiveness[t] == (
                    1 - exp(-blk.NTU[t] / blk.channels_per_pass * (1 + blk.Cratio[t]))
                ) / (1 + blk.Cratio[t])

        self.effectiveness_correlation = Constraint(
            self.flowsheet().time,
            rule=rule_Ecf,
            doc="Correlation for effectiveness factor",
        )

        # ---------------------------------------------------------------------
        # Pressure drop correlations
        # Friction factor calculation
        self.friction_factor_param_a = Param(
            initialize=0.0,
            units=pyunits.dimensionless,
            doc="Friction factor parameter A",
            mutable=True,
        )
        self.friction_factor_param_b = Param(
            initialize=18.29,
            units=pyunits.dimensionless,
            doc="Friction factor parameter B",
            mutable=True,
        )
        self.friction_factor_param_c = Param(
            initialize=-0.652,
            units=pyunits.dimensionless,
            doc="Friction factor parameter C",
            mutable=True,
        )

        def rule_fric_h(blk, t):
            return (
                blk.friction_factor_param_a
                + blk.friction_factor_param_b
                * blk.Re_hot[t] ** (blk.friction_factor_param_c)
            )

        self.friction_factor_hot = Expression(
            self.flowsheet().time, rule=rule_fric_h, doc="Hot side friction factor"
        )

        def rule_fric_c(blk, t):
            return (
                blk.friction_factor_param_a
                + blk.friction_factor_param_b
                * blk.Re_cold[t] ** (blk.friction_factor_param_c)
            )

        self.friction_factor_cold = Expression(
            self.flowsheet().time, rule=rule_fric_c, doc="Cold side friction factor"
        )

        def rule_hotside_dP(blk, t):
            return blk.hot_side.deltaP[t] == -(
                (
                    2
                    * blk.friction_factor_hot[t]
                    * (blk.plate_length + blk.port_diameter)
                    * blk.number_of_passes
                    * blk.hot_channel_velocity[t] ** 2
                    * blk.hot_side.properties_in[t].dens_mass
                    / blk.channel_diameter
                )
                + (
                    0.7
                    * blk.number_of_passes
                    * blk.hot_port_velocity[t] ** 2
                    * blk.hot_side.properties_in[t].dens_mass
                )
                + (
                    blk.hot_side.properties_in[t].dens_mass
                    * pyunits.convert(
                        Constants.acceleration_gravity,
                        to_units=units_meta("acceleration"),
                    )
                    * (blk.plate_length + blk.port_diameter)
                )
            )

        self.hot_side_deltaP_eq = Constraint(
            self.flowsheet().time, rule=rule_hotside_dP
        )

        def rule_coldside_dP(blk, t):
            return blk.cold_side.deltaP[t] == -(
                (
                    2
                    * blk.friction_factor_cold[t]
                    * (blk.plate_length + blk.port_diameter)
                    * blk.number_of_passes
                    * blk.cold_channel_velocity[t] ** 2
                    * pyunits.convert(
                        blk.cold_side.properties_in[t].dens_mass,
                        to_units=units_meta("density_mass"),
                    )
                    / blk.channel_diameter
                )
                + (
                    0.7
                    * blk.number_of_passes
                    * blk.cold_port_velocity[t] ** 2
                    * pyunits.convert(
                        blk.cold_side.properties_in[t].dens_mass,
                        to_units=units_meta("density_mass"),
                    )
                )
                + (
                    pyunits.convert(
                        blk.cold_side.properties_in[t].dens_mass,
                        to_units=units_meta("density_mass"),
                    )
                    * pyunits.convert(
                        Constants.acceleration_gravity,
                        to_units=units_meta("acceleration"),
                    )
                    * (blk.plate_length + blk.port_diameter)
                )
            )

        self.cold_side_deltaP_eq = Constraint(
            self.flowsheet().time, rule=rule_coldside_dP
        )

    def initialize(
        self,
        hot_side_state_args=None,
        cold_side_state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
        duty=None,
    ):
        """
        Heat exchanger initialization method.

        Args:
            hot_side_state_args : a dict of arguments to be passed to the
                property initialization for the hot side (see documentation of
                the specific property package) (default = None).
            cold_side_state_args : a dict of arguments to be passed to the
                property initialization for the cold side (see documentation of
                the specific property package) (default = None).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)
            duty : an initial guess for the amount of heat transferred. This
                should be a tuple in the form (value, units), (default
                = (1000 J/s))

        Returns:
            None

        """
        # Set solver options
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        hot_side = self.hot_side
        cold_side = self.cold_side

        # Create solver
        opt = get_solver(solver, optarg)

        flags1 = hot_side.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=hot_side_state_args
        )

        init_log.info_high("Initialization Step 1a (hot side) Complete.")

        flags2 = cold_side.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=cold_side_state_args
        )

        init_log.info_high("Initialization Step 1b (cold side) Complete.")

        # ---------------------------------------------------------------------
        # Solve unit without heat transfer equation
        self.energy_balance_constraint.deactivate()
        self.effectiveness_correlation.deactivate()
        self.effectiveness.fix(0.68)

        # Get side 1 and side 2 heat units, and convert duty as needed
        s1_units = hot_side.heat.get_units()
        s2_units = cold_side.heat.get_units()

        if duty is None:
            # Assume 1000 J/s and check for unitless properties
            if s1_units is None and s2_units is None:
                # Backwards compatibility for unitless properties
                s1_duty = -1000
                s2_duty = 1000
            else:
                s1_duty = pyunits.convert_value(
                    -1000, from_units=pyunits.W, to_units=s1_units
                )
                s2_duty = pyunits.convert_value(
                    1000, from_units=pyunits.W, to_units=s2_units
                )
        else:
            # Duty provided with explicit units
            s1_duty = -pyunits.convert_value(
                duty[0], from_units=duty[1], to_units=s1_units
            )
            s2_duty = pyunits.convert_value(
                duty[0], from_units=duty[1], to_units=s2_units
            )

        cold_side.heat.fix(s2_duty)
        for i in hot_side.heat:
            hot_side.heat[i].value = s1_duty

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)

        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))

        cold_side.heat.unfix()
        self.energy_balance_constraint.activate()

        for t in self.effectiveness:
            calculate_variable_from_constraint(
                self.effectiveness[t], self.effectiveness_correlation[t]
            )

        # ---------------------------------------------------------------------
        # Solve unit with new effectiveness factor
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))

        self.effectiveness_correlation.activate()
        self.effectiveness.unfix()

        # ---------------------------------------------------------------------
        # Final solve of full modelr
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 4 {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        # Release Inlet state
        hot_side.release_state(flags1, outlvl=outlvl)
        cold_side.release_state(flags2, outlvl=outlvl)

        init_log.info("Initialization Completed, {}".format(idaeslog.condition(res)))



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Python interface for ALAMO tool.
"""

from enum import Enum
import subprocess
from io import StringIO
import sys
import os
import json

import numpy as np
import pandas as pd

from pyomo.environ import Constraint, Expression, value, sin, cos, log, exp, Set, Reals
from pyomo.common.config import ConfigValue, In, Path, ListOf, Bool
from pyomo.common.tee import TeeStream
from pyomo.common.fileutils import Executable
from pyomo.common.tempfiles import TempfileManager

from idaes.core.surrogate.base.surrogate_base import SurrogateTrainer, SurrogateBase
from idaes.core.util.exceptions import ConfigurationError
import idaes.logger as idaeslog


# Set up logger
_log = idaeslog.getLogger(__name__)

# TODO: Adaptive sampling

alamo = Executable("alamo")

# Define mapping of Pyomo function names for expression evaluation
GLOBAL_FUNCS = {"sin": sin, "cos": cos, "ln": log, "exp": exp}


# The values associated with these must match those expected in the .alm file
class Modelers(Enum):
    """
    Enum for modelers supported by ALAMO.
    """

    BIC = 1
    MallowsCp = 2
    AICc = 3
    HQC = 4
    MSE = 5
    SSEP = 6
    RIC = 7
    MADp = 8


class Screener(Enum):
    """
    Enum for screeners supported by ALAMO.
    """

    none = 0
    lasso = 1
    SIS = 2


supported_options = [
    "xfactor",
    "xscaling",
    "scalez",
    "monomialpower",
    "multi2power",
    "multi3power",
    "ratiopower",
    "expfcns",
    "linfcns",
    "logfcns",
    "sinfcns",
    "cosfcns",
    "constant",
    # "grbfcns",  # TODO: deprecated
    # "rbfparam",  # TODO: deprecated
    "modeler",
    "builder",
    "backstepper",
    "convpen",
    "screener",
    "ncvf",
    "sismult",
    "maxtime",
    "maxiter",
    "datalimitterms",
    "maxterms",
    "minterms",
    "numlimitbasis",
    "exclude",
    "ignore",
    "xisint",
    "zisint",
    "tolrelmetric",
    "tolabsmetric",
    "tolmeanerror",
    "tolsse",
    "mipoptca",
    "mipoptcr",
    "linearerror",
    "GAMS",
    "GAMSSOLVER",
    "solvemip",
    "print_to_screen",
]

"""
Excluded Options:
NSAMPLE, NVALSAMPLE, INITIALIZER, PRINT TO FILE, FUNFORM, NTRANS

Excluded Simulator options:
MAXSIM, MINPOINTS, MAXPOINTS, SAMPLER, SIMULATOR, PRESET, TOLMAXERROR, SIMIN,
SIMOUT
"""


# Headers from ALAMO trace file that should be common for all outputs
common_trace = [
    "filename",
    "NINPUTS",
    "NOUTPUTS",
    "INITIALPOINTS",
    "SET",
    "INITIALIZER",
    "SAMPLER",
    "MODELER",
    "BUILDER",
    "GREEDYBUILD",
    "BACKSTEPPER",
    "GREEDYBACK",
    "REGULARIZER",
    "SOLVEMIP",
]


class AlamoTrainer(SurrogateTrainer):
    """
    Standard SurrogateTrainer for ALAMO.

    This defines a set of configuration options for ALAMO along with
    methods to read and write the ALAMO input and output files and to call
    the ALAMO executable.

    Generally, options default to None to indicate that no entry will be
    written in the ALAMO input file. In this case, the default ALAMO settings
    will be used.
    """

    # The following ALAMO options are not (yet) supported
    # Returning model prediction, as we can do that in IDAES
    # Iniital sampling of data, as SurrogateModelTrainer should do that
    # Similarly, adaptive sampling is better handled in Python
    # Single validation set, due to limitations of current API
    # Custom basis functions are not yet implemented

    CONFIG = SurrogateTrainer.CONFIG()

    CONFIG.declare(
        "xfactor",
        ConfigValue(
            default=None,
            domain=ListOf(float),
            description="List of scaling factors for input variables.",
        ),
    )
    CONFIG.declare(
        "xscaling",
        ConfigValue(
            default=None,
            domain=Bool,
            description="Option to scale input variables.",
            doc="Option to scale input variables. If True and xfactors are not "
            "provided, ALAMO sets XFACTORS equal to the range of each input "
            "variable.",
        ),
    )
    CONFIG.declare(
        "scalez",
        ConfigValue(
            default=None, domain=Bool, description="Option to scale output variables."
        ),
    )

    # Basis function options
    CONFIG.declare(
        "monomialpower",
        ConfigValue(
            default=None,
            domain=ListOf(
                int, In(Reals - {0, 1})
            ),  # allow any float except for 0 and 1
            description="Vector of monomial powers considered in basis "
            "functions - cannot include 0 or 1.",
        ),
    )
    CONFIG.declare(
        "multi2power",
        ConfigValue(
            default=None,
            domain=ListOf(float),
            description="Vector of powers to be considered for pairwise "
            "combinations in basis functions.",
        ),
    )
    CONFIG.declare(
        "multi3power",
        ConfigValue(
            default=None,
            domain=ListOf(float),
            description="Vector of three variable combinations of powers to be "
            "considered as basis functions.",
        ),
    )
    CONFIG.declare(
        "ratiopower",
        ConfigValue(
            default=None,
            domain=ListOf(float),
            description="Vector of ratio combinations of powers to be considered "
            "in the basis functions.",
        ),
    )
    CONFIG.declare(
        "constant",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Include constant basis function if True. Default = True",
        ),
    )
    CONFIG.declare(
        "linfcns",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Include linear basis functions if True. Default = True",
        ),
    )
    CONFIG.declare(
        "expfcns",
        ConfigValue(
            default=None,
            domain=Bool,
            description="Include exponential basis functions if True.",
        ),
    )
    CONFIG.declare(
        "logfcns",
        ConfigValue(
            default=None,
            domain=Bool,
            description="Include logarithmic basis functions if True.",
        ),
    )
    CONFIG.declare(
        "sinfcns",
        ConfigValue(
            default=None,
            domain=Bool,
            description="Include sine basis functions if True.",
        ),
    )
    CONFIG.declare(
        "cosfcns",
        ConfigValue(
            default=None,
            domain=Bool,
            description="Include cosine basis functions if True.",
        ),
    )
    CONFIG.declare(  # TODO: deprecated option
        "grbfcns",
        ConfigValue(
            default=None,
            domain=Bool,
            description="Include Gaussian radial basis functions if True.",
        ),
    )
    CONFIG.declare(  # TODO: deprecated option
        "rbfparam",
        ConfigValue(
            default=None,
            domain=float,
            description="Multiplicative constant used in the Gaussian radial basis"
            " functions.",
        ),
    )
    CONFIG.declare(
        "custom_basis_functions",
        ConfigValue(
            default=None,
            domain=ListOf(str),
            description="List of custom basis functions to include in surrogate "
            "fitting.",
            doc="List of custom basis functions to include in surrogate model "
            "fitting. These should be in a form that can be rendered as a string "
            "that meets ALAMO's requirements.",
        ),
    )

    # Other fitting options
    CONFIG.declare(
        "modeler",
        ConfigValue(
            default=None,
            domain=In(Modelers),
            description="Fitness metric to be used for model building. Must be an "
            "instance of Modelers Enum.",
        ),
    )
    CONFIG.declare(
        "builder",
        ConfigValue(
            default=None,
            domain=Bool,
            description="If True, a greedy heuristic builds up a model "
            "by adding one variable at a time.",
            doc="If True, a greedy heuristic builds up a model by adding one "
            "variable at a time. This model is used as a starting point for "
            "solving an integer programming formulation according to the choice "
            "of modeler. If an optimizer is not available, the heuristic model "
            "will be the final model to be returned.",
        ),
    )
    CONFIG.declare(
        "backstepper",
        ConfigValue(
            default=None,
            domain=Bool,
            description="If set to 1, a greedy heuristic builds down a model by "
            "starting from the least squares model and removing one variable at "
            "a time.",
        ),
    )
    CONFIG.declare(
        "convpen",
        ConfigValue(
            default=None,
            domain=float,
            description="Convex penalty term to use if Modeler == SSEP or MADp.",
            doc="When MODELER is set to 6 or 8, a penalty consisting of the sum "
            "of square errors (SSEP) or the maximum absolute error (MADp) and a "
            "term penalizing model size is used for model building. The size of "
            "the model is weighted by convpen. If convpen=0, this metric reduces "
            "to the classical sum of square errors (SSEP) or the maximum absolute "
            "deviation (MADp).",
        ),
    )
    CONFIG.declare(
        "screener",
        ConfigValue(
            default=None,
            domain=In(Screener),
            description="Regularization method used to reduce the number of "
            "potential basis functions before optimization. Must be instance of "
            "Screener Enum.",
        ),
    )
    CONFIG.declare(
        "ncvf",
        ConfigValue(
            default=None,
            domain=int,
            description="Number of folds to be used for cross validation by the "
            "lasso screener. ALAMO will use a two-fold validation if fewer than "
            "10 data points are available. NCVF must be a nonnegative integer.",
        ),
    )
    CONFIG.declare(
        "sismult",
        ConfigValue(
            default=None,
            domain=int,
            description="This parameter must be non-negative and is used to "
            "determine the number of basis functions retained by the SIS "
            "screener. The number of basis functions retained equals the floor "
            "of SSISmult n ln(n), where n is the number of measurements "
            "available at the current ALAMO iteration.",
        ),
    )

    CONFIG.declare(
        "maxiter",
        ConfigValue(
            default=None,
            domain=int,
            description="Maximum number of ALAMO iterations. 1 = no adaptive "
            "sampling, 0 = no limit.",
            doc="Maximum number of ALAMO iterations. Each iteration begins with "
            "a model-building step. An adaptive sampling step follows if maxiter "
            "does not equal 1. If maxiter is set to a number less than or equal "
            "to 0, ALAMO will enforce no limit on the number of iterations.",
        ),
    )
    CONFIG.declare(
        "maxtime",
        ConfigValue(
            default=1000,
            domain=float,
            description="Maximum total execution time allowed in seconds. "
            "Default = 1000.",
            doc="Maximum total execution time allowed in seconds. This time "
            "includes all steps of the algorithm, including time to read problem, "
            "preprocess data, solve optimization subproblems, and print results.",
        ),
    )
    CONFIG.declare(
        "datalimitterms",
        ConfigValue(
            default=None,
            domain=Bool,
            description="Limit model terms to number of measurements.",
            doc="If True, ALAMO will limit the number of terms in the model to be "
            "no more than the number of data measurements; otherwise, no limit "
            "based on the number of data measurements will be placed. The user "
            "may provide an additional limit on the number of terms in the model "
            "through the maxterms and minterms options.",
        ),
    )
    CONFIG.declare(
        "maxterms",
        ConfigValue(
            default=None,
            domain=ListOf(int),
            description="List of maximum number of model terms to per output.",
            doc="Row vector of maximum terms allowed in the modeling of output "
            "variables. One per output variable, space separated. A −1 signals "
            "that no limit is imposed.",
        ),
    )
    CONFIG.declare(
        "minterms",
        ConfigValue(
            default=None,
            domain=ListOf(int),
            description="List of minimum number of model terms to per output.",
            doc="Row vector of minimum terms required in the modeling of output "
            "variables. One per output variable, space separated. A 0 signals "
            "that no limit is imposed.",
        ),
    )
    CONFIG.declare(
        "numlimitbasis",
        ConfigValue(
            default=True,  # default this to true to avoid numerical issues
            domain=Bool,
            description="Eliminate infeasible basis functions. Default = True",
            doc="If True, ALAMO will eliminate basis functions that are not "
            "numerically acceptable (e.g., log(x) will be eliminated if x may be "
            "negative); otherwise, no limit based on the number of data "
            "measurements will be placed. The user may provide additional limits "
            "on the the type and number of selected basis functions through the "
            "options exclude and groupcon.",
        ),
    )
    CONFIG.declare(
        "exclude",
        ConfigValue(
            default=None,
            domain=ListOf(int),
            description="List of inputs to exclude during building,",
            doc="Row vector of 0/1 flags that specify which input variables, if "
            "any, ALAMO should exclude during the model building process. All "
            "input variables must be present in the data but ALAMO will not "
            "include basis functions that involve input variables for which "
            "exclude equals 1. This feature does not apply to custom basis "
            "functions or RBFs.",
        ),
    )
    CONFIG.declare(
        "ignore",
        ConfigValue(
            default=None,
            domain=ListOf(int),
            description="List of outputs to ignore during building.",
            doc="Row vector of 0/1 flags that specify which output variables, "
            "if any, ALAMO should ignore. All output variables must be present in "
            "the data but ALAMO does not model output variables for which ignore "
            "equals 1.",
        ),
    )
    CONFIG.declare(
        "xisint",
        ConfigValue(
            default=None,
            domain=ListOf(int),
            description="List of inputs that should be treated as integers.",
            doc="Row vector of 0/1 flags that specify which input variables, if "
            "any, ALAMO should treat as integers. For integer inputs, ALAMO’s "
            "sampling will be restricted to integer values.",
        ),
    )
    CONFIG.declare(
        "zisint",
        ConfigValue(
            default=None,
            domain=ListOf(int),
            description="List of outputs that should be treated as integers.",
            doc="Row vector of 0/1 flags that specify which output variables, if "
            "any, ALAMO should treat as integers. For integer variables, ALAMO’s "
            "model will include the rounding of a function to the nearest integer "
            "(equivalent to the nint function in Fortran.)",
        ),
    )

    CONFIG.declare(
        "tolrelmetric",
        ConfigValue(
            default=None,
            domain=ListOf(float),
            description="Relative tolerance for outputs.",
            doc="Relative convergence tolerance for the chosen fitness metric for "
            "the modeling of output variables. One per output variable, space "
            "separated. Incremental model building will stop if two consecutive "
            "iterations do not improve the chosen metric by at least this amount.",
        ),
    )
    CONFIG.declare(
        "tolabsmetric",
        ConfigValue(
            default=None,
            domain=ListOf(float),
            description="Absolute tolerance for outputs.",
            doc="Absolute convergence tolerance for the chosen fitness metric for "
            "the modeling of output variables. One per output variable, space "
            "separated. Incremental model building will stop if two consecutive "
            "iterations do not improve the chosen metric by at least this amount.",
        ),
    )
    CONFIG.declare(
        "tolmeanerror",
        ConfigValue(
            default=None,
            domain=ListOf(float),
            description="Convergence tolerance for mean errors in outputs.",
            doc="Row vector of convergence tolerances for mean errors in the "
            "modeling of output variables. One per output variable, space "
            "separated. Incremental model building will stop if tolmeanerror, "
            "tolrelmetric, or tolabsmetric is satisfied.",
        ),
    )
    CONFIG.declare(
        "tolsse",
        ConfigValue(
            default=None,
            domain=float,
            description="Absolute tolerance on SSE",
            doc="Absolute tolerance on sum of square errors (SSE). ALAMO will "
            "terminate if it finds a solution whose SSE is within tolsse from "
            "the SSE of the full least squares problem.",
        ),
    )

    CONFIG.declare(
        "mipoptca",
        ConfigValue(
            default=None, domain=float, description="Absolute tolerance for MIP."
        ),
    )
    CONFIG.declare(
        "mipoptcr",
        ConfigValue(
            default=None, domain=float, description="Relative tolerance for MIP."
        ),
    )
    CONFIG.declare(
        "linearerror",
        ConfigValue(
            default=None,
            domain=Bool,
            description="If True, a linear objective is used when solving "
            "mixed-integer optimization problems; otherwise, a squared error will "
            "be employed.",
        ),
    )
    CONFIG.declare(
        "GAMS",
        ConfigValue(
            default=None,
            domain=str,
            description="Complete path of GAMS executable (or name if GAMS is in "
            "the user path).",
        ),
    )
    CONFIG.declare(
        "GAMSSOLVER",
        ConfigValue(
            default=None,
            domain=str,
            description="Name of preferred GAMS solver for solving ALAMO’s "
            "mixed-integer quadratic subproblems. Special facilities have been "
            "implemented in ALAMO and BARON that make BARON the preferred "
            "selection for this option. However, any mixed-integer quadratic "
            "programming solver available under GAMS can be used.",
        ),
    )
    CONFIG.declare(
        "solvemip",
        ConfigValue(
            default=None,
            domain=Bool,
            description="Whether to use an optimizer to solve MIP.",
        ),
    )
    CONFIG.declare(
        "print_to_screen",
        ConfigValue(
            default=None,
            domain=Bool,
            description="Send ALAMO output to stdout. Output is returned by the "
            "call_alamo method.",
        ),
    )

    # I/O file options
    CONFIG.declare(
        "alamo_path",
        ConfigValue(
            default=None, domain=Path, doc="Path to ALAMO executable (if not in path)."
        ),
    )
    CONFIG.declare(
        "filename",
        ConfigValue(
            default=None,
            domain=str,
            description="File name to use for ALAMO files - must be full path of a"
            " .alm file. Other files will be defined from this pattern. If this "
            "option is not None, then working files will not be deleted.",
        ),
    )
    CONFIG.declare(
        "working_directory",
        ConfigValue(
            default=None,
            domain=str,
            description="Full path to working directory for ALAMO to use. "
            "If this option is not None, then working files will not be deleted.",
        ),
    )
    CONFIG.declare(
        "overwrite_files",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Flag indicating whether existing files can be " "overwritten.",
        ),
    )

    # TODO: We need to do some processing of the labels since ALAMO is
    # restrictive about the labels
    # TODO: We need to think more carefully about "input_bounds".
    # Alamo uses bounds during training, but we also want to consider
    # "valid" bounds for the surrogate
    def __init__(self, **settings):
        super().__init__(**settings)

        # TODO: Deprecation warnings for RBF options
        if self.config.grbfcns is not None or self.config.rbfparam is not None:
            raise ConfigurationError(
                "Support for radial basis functions in the ALAMOpy wrapper have been "
                "deprecated due to limitations in the current implementation. "
                "Pull requests are welcome to address this, otherwise we recommend using "
                "Pysmo which includes support for radial basis functions."
            )

        self._temp_context = None
        self._almfile = None
        self._trcfile = None

        if self.config.alamo_path is not None:
            alamo.executable = self.config.alamo_path

        self._results = None

    def train_surrogate(self):
        """
        General workflow method for training an ALAMO surrogate.

        Takes the existing data set and executes the ALAMO workflow to create
        an AlamoSurrogate based on the current configuration arguments.

        Args:
            None

        Returns:
            tuple : (success, AlamoSurrogate, message) where success indicates
            whether ALAMO was usccessfully executed, an instance of an
            AlamoSurrogate representing the trained surrogate, and message is
            the final status line from the ALAMO output log.
        """
        # Get paths for temp files
        self._get_files()

        return_code = None
        alamo_log = None
        alamo_object = None

        # Check for any issues in arguments that might cause problems
        self._verify_inputs()

        try:
            # Write .alm file
            self._write_alm_file()

            # Call ALAMO executable
            return_code, alamo_log = self._call_alamo()

            # Read back results
            trace_dict = self._read_trace_file(self._trcfile)

            # Populate results and SurrogateModel object
            self._populate_results(trace_dict)
            alamo_object = self._build_surrogate_object()

        finally:
            # Clean up temporary files if required
            self._remove_temp_files()

        success = False
        if return_code == 0 and alamo_object is not None:
            # Non-zero return code implies an error
            # specifics returned in the msg
            success = True
        alamo_msg = alamo_log.split("\n")[-3]

        return success, alamo_object, alamo_msg

    # TODO: let's generalize this under the metrics?
    def get_alamo_results(self):
        """Return results from trainer"""
        return self._results

    def _get_files(self):
        """
        Method to get/set paths for .alm and .trc files based on filename
        configuration argument.

        If filename is None, temporary files will be created.

        Args:
            None

        Returns:
            None
        """
        if self._temp_context is None:
            self._temp_context = TempfileManager.new_context()

        if self.config.filename is None:
            # Get a temporary file from the manager
            almfile = self._temp_context.create_tempfile(suffix=".alm")
        else:
            almfile = self.config.filename

            if not self.config.overwrite_files:
                # It is OK if the trace file exists, as ALAMO will append to it
                # The trace file reader handles this case
                if os.path.isfile(almfile):
                    raise FileExistsError(
                        f"A file with the name {almfile} already exists. "
                        f"Either choose a new file name or set "
                        f"overwrite_files = True."
                    )

            self._temp_context.add_tempfile(almfile, exists=False)

        trcfile = os.path.splitext(almfile)[0] + ".trc"
        self._temp_context.add_tempfile(trcfile, exists=False)

        if self.config.working_directory is None:
            wrkdir = self._temp_context.create_tempdir()
        else:
            wrkdir = self.config.working_directory

        # Set attributes to track file names
        self._almfile = almfile
        self._trcfile = trcfile
        self._wrkdir = wrkdir

    def _verify_inputs(self):
        """
        Check for issues in input arguments that will cause issues later.
        """
        # ALAMO does not support spaces in variable names
        # TODO: Should this be a check for all IDAES surrogate models?
        for i in self._input_labels:
            if " " in i:
                raise ValueError(
                    f"ALAMO does not support the presence of spaces in variable names. "
                    f"Found space in input names: {i}."
                )
        for o in self._output_labels:
            if " " in o:
                raise ValueError(
                    f"ALAMO does not support the presence of spaces in variable names. "
                    f"Found space in output names: {o}."
                )

    def _write_alm_to_stream(
        self, stream, trace_fname=None, training_data=None, validation_data=None
    ):
        """
        Method to write an ALAMO input file (.alm) to a stream.
        Users may provide specific data sets for training and validation.
        If no data sets are provided, the data sets contained in the
        AlamoModelTrainer are used.

        Args:
            stream: stream that data should be written to
            trace_fname: name for trace file (.trc) to be included in .alm file
            training_data: Pandas dataframe to use for training surrogate
            validation_data: Pandas dataframe to use for validating surrogate

        Returns:
            None
        """
        if training_data is None:
            training_data = self._training_dataframe
        if validation_data is None:
            validation_data = self._validation_dataframe

        # Check bounds on inputs to avoid potential ALAMO failures
        input_max = list()
        input_min = list()
        for k, b in self._input_bounds.items():
            if b is None or b[0] is None or b[1] is None:
                raise ConfigurationError(
                    f"ALAMO configuration error: invalid bounds on input {k} " f"({b})."
                )
            elif b[0] == b[1]:
                raise ConfigurationError(
                    f"ALAMO configuration error: upper and lower bounds on "
                    f"input {k} are equal."
                )
            elif b[1] < b[0]:
                raise ConfigurationError(
                    f"ALAMO configuration error: upper bound is less than "
                    f"lower bound for input {k}."
                )
            input_min.append(b[0])
            input_max.append(b[1])

        # Get number of data points to build alm file
        n_rdata, n_inputs = training_data.shape  # pylint: disable=unused-variable

        if validation_data is not None:
            n_vdata, n_inputs = validation_data.shape
        else:
            n_vdata = 0

        stream.write("# IDAES Alamopy input file\n")
        stream.write(f"NINPUTS {len(self._input_labels)}\n")
        stream.write(f"NOUTPUTS {len(self._output_labels)}\n")
        stream.write(f"XLABELS {' '.join(map(str, self._input_labels))}\n")
        stream.write(f"ZLABELS {' '.join(map(str, self._output_labels))}\n")
        stream.write(f"XMIN {' '.join(map(str, input_min))}\n")
        stream.write(f"XMAX {' '.join(map(str, input_max))}\n")
        stream.write(f"NDATA {n_rdata}\n")
        if validation_data is not None:
            stream.write(f"NVALDATA {n_vdata}\n")
        stream.write("\n")

        # Other options for config
        # Can be bool, list of floats, None, Enum, float
        # Special cases
        if self.config.monomialpower is not None:
            stream.write(f"MONO {len(self.config.monomialpower)}\n")
        if self.config.multi2power is not None:
            stream.write(f"MULTI2 {len(self.config.multi2power)}\n")
        if self.config.multi3power is not None:
            stream.write(f"MULTI3 {len(self.config.multi3power)}\n")
        if self.config.ratiopower is not None:
            stream.write(f"RATIOS {len(self.config.ratiopower)}\n")
        if self.config.custom_basis_functions is not None:
            stream.write(f"NCUSTOMBAS {len(self.config.custom_basis_functions)}\n")

        for o in supported_options:
            if self.config[o] is None:
                # Write nothing to alm file
                continue
            elif isinstance(self.config[o], Enum):
                # Need to write value of Enum
                stream.write(f"{o} {self.config[o].value}\n")
            elif isinstance(self.config[o], bool):
                # Cast bool to int
                stream.write(f"{o} {int(self.config[o])}\n")
            elif isinstance(self.config[o], (str, float, int)):
                # Write value to file
                stream.write(f"{o} {self.config[o]}\n")
            else:
                # Assume the argument is a list
                stream.write(f"{o} {' '.join(map(str, self.config[o]))}\n")

        stream.write("\nTRACE 1\n")
        if trace_fname is not None:
            stream.write(f"TRACEFNAME {trace_fname}\n")

        def _df_to_data_fragment(df, **kwargs):
            text = df.to_string(
                header=False,
                index=False,
                float_format=lambda x: str(x).format(":g"),
                **kwargs,
            )
            return text

        stream.write("\nBEGIN_DATA\n")
        # Columns will be written in order in input and output lists
        training_data_str = _df_to_data_fragment(
            training_data,
            columns=self._input_labels + self._output_labels,
        )
        stream.write(training_data_str)
        stream.write("\nEND_DATA\n")

        if validation_data is not None:
            # Add validation data definition
            stream.write("\nBEGIN_VALDATA\n")
            val_data_str = _df_to_data_fragment(
                validation_data,
                columns=self._input_labels + self._output_labels,
            )
            stream.write(val_data_str)
            stream.write("\nEND_VALDATA\n")

        if self.config.custom_basis_functions is not None:
            stream.write("\nBEGIN_CUSTOMBAS\n")
            for i in self.config.custom_basis_functions:
                stream.write(f"{str(i)}\n")
            stream.write("END_CUSTOMBAS\n")

    def _write_alm_file(self, training_data=None, validation_data=None):
        """
        Method to write an ALAMO input file (.alm) using the current settings.
        Users may provide specific data sets for training and validation.
        If no data sets are provided, the data sets contained in the
        AlamoModelTrainer are used.

        Args:
            training_data: Pandas dataframe to use for training surrogate
            validation_data: Pandas dataframe to use for validating surrogate

        Returns:
            None
        """
        f = open(self._almfile, "w")
        self._write_alm_to_stream(f, self._trcfile, training_data, validation_data)
        f.close()

    def _call_alamo(self):
        """
        Method to call ALAMO executable from Python, passing the current .alm
        file as an argument.

        Args:
            None

        Returns:
            ALAMO: return code
            log: string of the text output from ALAMO
        """
        if alamo.executable is None:
            raise FileNotFoundError(
                "Could not find ALAMO executable. Please ensure that ALAMO "
                "is installed and in the system path, or provide a path to "
                "the executable."
            )

        ostreams = [StringIO(), sys.stdout]

        if self._temp_context is None:
            self._get_files()

        # Set working directory
        cwd = os.getcwd()
        os.chdir(self._wrkdir)

        try:
            # Add lst file to temp file manager
            lstfname = os.path.splitext(os.path.basename(self._almfile))[0] + ".lst"
            lstpath = os.path.join(cwd, lstfname)
            self._temp_context.add_tempfile(lstpath, exists=False)

            with TeeStream(*ostreams) as t:
                results = subprocess.run(
                    [alamo.executable, str(self._almfile)],
                    stdout=t.STDOUT,
                    stderr=t.STDERR,
                    universal_newlines=True,
                    check=False,
                )

                t.STDOUT.flush()
                t.STDERR.flush()

            return_code = results.returncode
            alamo_log = ostreams[0].getvalue()

        except OSError:
            _log.error(
                f"Could not execute the command: alamo {str(self._almfile)}. ",
                f"Error message: {sys.exc_info()[1]}.",
            )
            raise

        finally:
            # Revert cwd to where it started
            os.chdir(cwd)

        if "ALAMO terminated with termination code " in alamo_log:
            self._remove_temp_files()
            _log.warning(
                "ALAMO executable returned non-zero return code. Check "
                "the ALAMO output for more information."
            )

        return return_code, alamo_log

    def _read_trace_file(self, trcfile, has_validation_data=False):
        """
        Method to read the results of an ALAMO run from a trace (.trc) file.
        The name location of the trace file is tored on the AlamoModelTrainer
        object and is generally set automatically.

        Args:
            trcfile : str
               Path to the trcfile to read
            output_labels : list of str
               List of strings of the output_labels (in order)
            has_validation_data : bool
                Bool indicating whether valdiation data was included in ALAMO run

        Returns:
            trace_dict: contents of trace file as a dict
        """
        try:
            with open(trcfile, "r") as f:
                lines = f.readlines()
            f.close()
        except FileNotFoundError:
            # Trace file does not exist
            raise FileNotFoundError(
                "Error occurred when trying to read the ALAMO trace file - this probably "
                "indicates that a trace file was not created by the ALAMO executable. "
                "Please check the ALAMO output logs."
            )

        output_labels = self.output_labels()

        trace_read = {}
        # Get headers from first line in trace file
        headers = lines[0].split(", ")
        for i, v in enumerate(headers):
            header = v.strip("#\n")
            if header in common_trace:
                trace_read[header] = None
            else:
                trace_read[header] = {}

        # Get trace output from final line(s) of file
        # ALAMO will append new lines to existing trace files
        # For multiple outputs, each output has its own line in trace file
        if self._validation_dataframe is not None or has_validation_data:
            omult = 2
        else:
            omult = 1

        for j, ol in enumerate(output_labels):
            trace = lines[(-len(output_labels) + j) * omult].split(", ")

            for i, v in enumerate(headers):
                header = v.strip("#\n")
                trace_val = trace[i].strip("\n")

                # Replace Fortran powers (^) with Python powers (**)
                trace_val = trace_val.replace("^", "**")
                # Replace = with ==
                trace_val = trace_val.replace("=", "==")

                if header in common_trace:
                    # These should be common across all output
                    if trace_read[header] is None:
                        # No value yet, so set value
                        trace_read[header] = trace_val
                    else:
                        # Check that current value matches the existing value
                        if trace_read[header] != trace_val:
                            raise RuntimeError(
                                f"Mismatch in values when reading ALAMO trace "
                                f"file. Values for {header}: "
                                f"{trace_read[header]}, {header}"
                            )
                else:
                    trace_read[header][ol] = trace_val

                # Do some final sanity checks
                if header == "OUTPUT":
                    # OUTPUT should be equal to the current index of outputs
                    if trace_val != str(j + 1):
                        raise RuntimeError(
                            f"Mismatch when reading ALAMO trace file. "
                            f"Expected OUTPUT = {j+1}, found {trace_val}."
                        )
                elif header == "SET":
                    # SET should always be 0 - higher numbers are for
                    # validation data sets
                    if trace_val != "0":
                        raise RuntimeError(
                            f"Mismatch when reading ALAMO trace file. "
                            f"Expected SET = 0, found {trace_val}. "
                            f"This likely indicates the presence of "
                            f"unexpected validation data sets."
                        )
                elif header == "Model":
                    # Var label on LHS should match output label
                    vlabel = trace_val.split("==")[0].strip()
                    if vlabel != ol:
                        raise RuntimeError(
                            f"Mismatch when reading ALAMO trace file. "
                            f"Label of output variable in expression "
                            f"({vlabel}) does not match expected label "
                            f"({ol})."
                        )

        return trace_read

    def _populate_results(self, trace_dict):
        """
        Method to populate the results object with data from a trace file.

        Args:
            trace_dict: trace file data in form of a dict

        Returns:
            None
        """
        self._results = trace_dict

    def _build_surrogate_object(self):
        """
        Method to construct an AlmaoObject from the current results
        object.

        Args:
            None

        Returns:
            AlamoSurrogate
        """
        return AlamoSurrogate(
            surrogate_expressions=self._results["Model"],
            input_labels=self._input_labels,
            output_labels=self._output_labels,
            input_bounds=self._input_bounds,
        )

    def _remove_temp_files(self):
        """
        Method to remove temporary files created during the ALAMO workflow,
        i.e. the .alm and .trc files.

        Args:
            None

        Returns:
            None
        """
        if self._temp_context is None:
            # Assume we have already cleaned up
            return

        remove = True
        if self.config.filename is not None:
            remove = False

        self._temp_context.release(remove=remove)
        # Release tempfile context
        self._temp_context = None


class AlamoSurrogate(SurrogateBase):
    """
    Standard SurrogateObject for surrogates trained using ALAMO.

    Contains methods to both populate a Pyomo Block with constraints
    representing the surrogate and to evaluate the surrogate a set of user
    provided points.
    """

    def __init__(
        self, surrogate_expressions, input_labels, output_labels, input_bounds=None
    ):
        super().__init__(input_labels, output_labels, input_bounds)
        self._surrogate_expressions = surrogate_expressions
        self._fcn = None

    def evaluate_surrogate(self, inputs):
        """
        Method to evaluate the ALAMO surrogate model at a set of user provided values.

        Args:
           dataframe: pandas DataFrame
              The dataframe of input values to be used in the evaluation. The dataframe
              needs to contain a column corresponding to each of the input labels. Additional
              columns are fine, but are not used.

        Returns:
            output: pandas Dataframe
              Returns a dataframe of the output values evaluated at the provided inputs.
              The index of the output dataframe should match the index of the provided inputs.
        """
        # Create a set of lambda functions for evaluating the surrogate.
        if self._fcn is None:
            fcn = dict()
            for o in self._output_labels:
                # We need to evaluate the string returned by ALAMO
                # pylint: disable=W0123
                fcn[o] = eval(
                    f"lambda {', '.join(self._input_labels)}: "
                    f"{self._surrogate_expressions[o].split('==')[1]}",
                    GLOBAL_FUNCS,
                )
            self._fcn = fcn

        # Use numpy to do the calculations as it is faster
        inputdata = inputs[self._input_labels].to_numpy()
        outputs = np.zeros(shape=(inputs.shape[0], len(self._output_labels)))

        for i in range(inputdata.shape[0]):
            for o, o_name in enumerate(self._output_labels):
                outputs[i, o] = value(self._fcn[o_name](*inputdata[i, :]))

        return pd.DataFrame(
            data=outputs, index=inputs.index, columns=self._output_labels
        )

    def populate_block(self, block, additional_options=None):
        """
        Method to populate a Pyomo Block with surrogate model constraints
        or expressions.

        Args:
            block: Pyomo Block component to be populated with constraints.
            additional_options (dict): optional keyword arguments passes along
                from build model. There is one optional argument, as_expression
                (bool) if True, write the model as an expressions, if False or
                not provided write model as constraints. If as_expression is
                True, the output_variables argument to build_model() is not
                required.
        Returns:
            None
        """

        # TODO: do we need to add the index_set stuff back in?
        output_set = Set(initialize=self._output_labels, ordered=True)

        if additional_options is not None:
            as_expression = additional_options.pop("as_expression", False)
        else:
            as_expression = False

        if not as_expression:

            def alamo_rule(b, o):
                lvars = block.input_vars_as_dict()
                lvars.update(block.output_vars_as_dict())
                # We need to evaluate the string returned by ALAMO
                # pylint: disable=W0123
                return eval(self._surrogate_expressions[o], GLOBAL_FUNCS, lvars)

            block.alamo_constraint = Constraint(output_set, rule=alamo_rule)
        else:

            def alamo_rule(b, o):
                lvars = block.input_vars_as_dict()
                # We need to evaluate the string returned by ALAMO
                # pylint: disable=W0123
                return eval(
                    self._surrogate_expressions[o].split("==")[1], GLOBAL_FUNCS, lvars
                )

            block.alamo_expression = Expression(output_set, rule=alamo_rule)

    def save(self, strm):
        """
        Save an instance of this surrogate to the strm so the model can be used later.

        Args:
           strm: IO.TextIO
              This is the python stream like a file object or StringIO that will be used
              to serialize the surrogate object. This method writes a string
              of json data to the stream.
        """
        json.dump(
            {
                "surrogate": self._surrogate_expressions,
                "input_labels": self._input_labels,
                "output_labels": self._output_labels,
                "input_bounds": self._input_bounds,
            },
            strm,
        )

    @classmethod
    def load(cls, strm):
        """
        Create an instance of a surrogate from a stream.

        Args:
           strm: stream
              This is the python stream containing the data required to load the surrogate.
              This is often, but does not need to be a string of json data.

        Returns: an instance of the derived class or None if it failed to load
        """
        d = json.load(strm)

        surrogate_expressions = d["surrogate"]
        input_labels = d["input_labels"]
        output_labels = d["output_labels"]

        # Need to convert list of bounds to tuples
        input_bounds = {}
        for k, v in d["input_bounds"].items():
            input_bounds[k] = tuple(v)

        return AlamoSurrogate(
            surrogate_expressions=surrogate_expressions,
            input_labels=input_labels,
            output_labels=output_labels,
            input_bounds=input_bounds,
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

"""
This is an example supercritical pulverized coal (SCPC) power plant, including
steam cycle and boiler heat exchanger network model.

This simulation model consist of a ~595 MW gross coal fired power plant.
The dimensions and operating conditions used for this simulation do not
represent any specific coal-fired power plant.

This model is for demonstration and tutorial purposes only.
Before looking at the model, it may be useful to look
at the process flow diagram (PFD).

SCPC Power Plant

Inputs:
    Fresh Water (water make up)
    Throttle valve opening,
    BFW - boiler feed water (from Feed water heaters)
    Coal from pulverizers

Main Assumptions:
    Coal flowrate as a function of load, coal HHV is fixed and heat dutty
    splitt from fire side to water wall and platen superheater is fixed.

    Boiler heat exchanger network:
        Water Flow:
            Fresh water -> FWH's -> Economizer -> Water Wall -> Primary SH -> Platen SH -> Finishing Superheate -> HP Turbine -> Reheater -> IP Turbine
        Flue Gas Flow:
            Fire Ball -> Platen SH -> Finishing SH -> Reheater  -> o -> Economizer -> Air Preheater
                                                   -> Primary SH --^
        Steam Flow:
            Boiler -> HP Turbine -> Reheater -> IP Turbine
            HP, IP, and LP steam extractions to Feed Water Heaters


    Models used:
        - Mixers: Attemperator, Flue gas mix
        - Heater: Platen SH, Fire/Water side (simplified model),
                  Feed Water Heaters, Hot Tank, Condenser
        - BoilerHeatExchanger: Economizer, Primary SH, Finishing SH, Reheater
            + Shell and tube heat exchanger
                - tube side: Steam (side 1 holdup)
                - shell side: flue gas (side 2 holdup)
        - Steam Turbines
        - Pumps
    Property packages used:
        - IAPWS: Water/steam side
        - IDEAL GAS: Flue Gas side

"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

__author__ = "Miguel Zamarripa"

# Import Python libraries
import logging

# Import Pyomo libraries
import pyomo.environ as pyo
from pyomo.network import Arc

# IDAES Imports
from idaes.core.util.model_statistics import degrees_of_freedom
import idaes.logger as idaeslog

_log = idaeslog.getModelLogger(__name__, logging.INFO)


def import_steam_cycle():
    # build concrete model
    # import steam cycle model and initialize flowsheet (only if needed)
    # pylint: disable-next=import-outside-toplevel
    import idaes.models_extra.power_generation.flowsheets.supercritical_steam_cycle.supercritical_steam_cycle as steam_cycle

    m, solver = steam_cycle.main()
    return m, solver


def main():
    # import steam cycle and build concrete model
    m, solver = import_steam_cycle()
    print(degrees_of_freedom(m))
    # at this point we have a flowsheet with "steam cycle" that solves
    # correctly, with 0 degrees of freedom.

    # next step is to import and build the boiler heat exchanger network
    # importing the boiler heat exchanger network
    # from (boiler_subflowsheet_build.py)
    # this step appends all the boiler unit models into our model ("m")
    # model "m" has been created a few lines above
    # pylint: disable-next=import-outside-toplevel
    import idaes.models_extra.power_generation.flowsheets.supercritical_power_plant.boiler_subflowsheet_build as blr

    # import the models (ECON, WW, PrSH, PlSH, FSH, Splitter, Mixer, Reheater)
    # see boiler_subflowhseet_build.py for a better description
    blr.build_boiler(m.fs)
    # initialize boiler network models (one by one)
    blr.initialize(m)
    # at this point we have both flowsheets (steam cycle + boiler network)
    # in the same model/concrete object ("m"), however they are disconnected.
    # Here we want to solve them at the same time
    # this is a square problem (i.e. degrees of freedom = 0)
    print("solving square problem disconnected")
    results = solver.solve(m, tee=True)

    # at this point we want to connect the units in both flowsheets
    # Economizer inlet = Feed water heater 8 outlet (water)
    # HP inlet = Attemperator outlet (steam)
    # Reheater inlet (steam) = HP split 7 outlet (last stage of HP turbine)
    # IP inlet = Reheater outlet steam7
    blr.unfix_inlets(m)
    print("unfix inlet conditions, degreeso of freedom = " + str(degrees_of_freedom(m)))
    # user can save the initialization to a json file (uncomment next line)
    #    MS.to_json(m, fname = 'SCPC_full.json')
    #   later user can use the json file to initialize the model
    #   if this is the case comment out previous MS.to_json and uncomment next line
    #    MS.from_json(m, fname = 'SCPC_full.json')

    # deactivate constraints linking the FWH8 to HP turbine
    m.fs.boiler_pressure_drop.deactivate()
    m.fs.close_flow.deactivate()
    m.fs.turb.constraint_reheat_flow.deactivate()
    m.fs.turb.constraint_reheat_press.deactivate()
    m.fs.turb.constraint_reheat_temp.deactivate()
    m.fs.turb.inlet_split.inlet.enth_mol.unfix()
    m.fs.turb.inlet_split.inlet.pressure.unfix()
    # user can fix the boiler feed water pump pressure (uncommenting next line)
    #    m.fs.bfp.outlet.pressure[:].fix(26922222.222))

    m.fs.FHWtoECON = Arc(
        source=m.fs.fwh8.desuperheat.cold_side_outlet,
        destination=m.fs.ECON.cold_side_inlet,
    )

    m.fs.Att2HP = Arc(source=m.fs.ATMP1.outlet, destination=m.fs.turb.inlet_split.inlet)

    m.fs.HPout2RH = Arc(
        source=m.fs.turb.hp_split[7].outlet_1, destination=m.fs.RH.cold_side_inlet
    )

    m.fs.RHtoIP = Arc(
        source=m.fs.RH.cold_side_outlet, destination=m.fs.turb.ip_stages[1].inlet
    )

    pyo.TransformationFactory("network.expand_arcs").apply_to(m)

    # unfix boiler connections
    m.fs.ECON.cold_side_inlet.flow_mol.unfix()
    m.fs.ECON.cold_side_inlet.enth_mol[0].unfix()
    m.fs.ECON.cold_side_inlet.pressure[0].unfix()
    m.fs.RH.cold_side_inlet.flow_mol.unfix()
    m.fs.RH.cold_side_inlet.enth_mol[0].unfix()
    m.fs.RH.cold_side_inlet.pressure[0].unfix()
    m.fs.hotwell.makeup.flow_mol[:].setlb(-1.0)

    # if user has trouble with infeasible solutions, an easy test
    # is to deactivate the link to HP turbine
    # (m.fs.Att2HP_expanded "enth_mol and pressure" equalities)
    # and fix inlet pressure and enth_mol to turbine
    # (m.fs.turb.inlet_split.inlet)
    # (then double check the values from m.fs.ATMP1.outlet)
    #  m.fs.Att2HP_expanded.enth_mol_equality.deactivate()
    #  m.fs.Att2HP_expanded.pressure_equality.deactivate()
    m.fs.turb.inlet_split.inlet.pressure.fix(2.423e7)
    #    m.fs.turb.inlet_split.inlet.enth_mol.fix(62710.01)

    # finally, since we want to maintain High Pressure (HP) inlet temperature
    # constant (~866 K), we need to fix Attemperator enthalpy outlet
    # and unfix heat duty to Platen superheater, note that fixing enthalpy
    # to control temperature is only valid because pressure is also fixed
    m.fs.ATMP1.outlet.enth_mol[0].fix(62710.01)
    m.fs.PlSH.heat_duty[:].unfix()  # fix(5.5e7)
    #    m.fs.ATMP1.SprayWater.flow_mol[0].unfix()
    print("connecting flowsheets, degrees of freedom = " + str(degrees_of_freedom(m)))
    print("solving full plant model")
    solver.options = {
        "tol": 1e-6,
        "linear_solver": "ma27",
        "max_iter": 40,
    }
    # square problems tend to work better without bounds
    strip_bounds = pyo.TransformationFactory("contrib.strip_var_bounds")
    strip_bounds.apply_to(m, reversible=True)
    # this is the final solve with both flowsheets connected
    results = solver.solve(m, tee=True)
    return m, results


if __name__ == "__main__":
    m, results = main()
